import os
import platform
import shlex
import shutil
import sys
from logging import getLogger
from pathlib import Path
from typing import List

from PyQt5.QtCore import QStandardPaths

if platform.system() == "Windows":
    # noinspection PyUnresolvedReferences
    from win32com.client import Dispatch  # pylint: disable=E0401

logger = getLogger("Paths")

resources_path = Path(__file__).absolute().parent.parent.joinpath("resources")

# lk: delete old Rare directories
for old_dir in [
    Path(QStandardPaths.writableLocation(QStandardPaths.CacheLocation), "rare").joinpath("tmp"),
    Path(QStandardPaths.writableLocation(QStandardPaths.AppDataLocation), "rare").joinpath("images"),
    Path(QStandardPaths.writableLocation(QStandardPaths.CacheLocation), "rare"),
    Path(QStandardPaths.writableLocation(QStandardPaths.AppDataLocation), "rare"),
]:
    if old_dir.exists():
        # lk: case-sensitive matching on Winblows
        if old_dir.stem in old_dir.parent.iterdir():
            shutil.rmtree(old_dir, ignore_errors=True)


# lk: TempLocation doesn't depend on OrganizationName or ApplicationName
# lk: so it is fine to use it before initializing the QApplication
def lock_file() -> Path:
    return Path(QStandardPaths.writableLocation(QStandardPaths.TempLocation), "Rare.lock")


def config_dir() -> Path:
    # FIXME: This returns ~/.config/Rare/Rare/ for some reason while the settings are in ~/.config/Rare/Rare.conf
    # Take the parent for now, but this should be investigated
    return Path(QStandardPaths.writableLocation(QStandardPaths.AppConfigLocation)).parent


def data_dir() -> Path:
    return Path(QStandardPaths.writableLocation(QStandardPaths.AppDataLocation))


def cache_dir() -> Path:
    return Path(QStandardPaths.writableLocation(QStandardPaths.CacheLocation))


def image_dir() -> Path:
    return data_dir().joinpath("images")


def log_dir() -> Path:
    return cache_dir().joinpath("logs")


def tmp_dir() -> Path:
    return cache_dir().joinpath("tmp")


def create_dirs() -> None:
    for path in (data_dir(), cache_dir(), image_dir(), log_dir(), tmp_dir()):
        if not path.exists():
            path.mkdir(parents=True)
            logger.info(f"Created directory at {path}")


def home_dir() -> Path:
    return Path(QStandardPaths.writableLocation(QStandardPaths.HomeLocation))


def desktop_dir() -> Path:
    return Path(QStandardPaths.writableLocation(QStandardPaths.DesktopLocation))


def applications_dir() -> Path:
    return Path(QStandardPaths.writableLocation(QStandardPaths.ApplicationsLocation))

# fmt: off
__link_suffix = {
    "Windows": {
        "link": "lnk",
        "icon": "ico",
    },
    "FreeBSD": {
        "link": "desktop",
        "icon": "png",
    },
    "Linux": {
        "link": "desktop",
        "icon": "png",
    },
    "Darwin": {
         "link": "",
         "icon": "icns",
    },
}

def desktop_links_supported() -> bool:
    supported_systems = [k for (k, v) in __link_suffix.items() if v["link"]]
    return platform.system() in supported_systems

def desktop_icon_suffix() -> str:
    return __link_suffix[platform.system()]["icon"]


__link_type = {
    "desktop": desktop_dir(),
    # lk: for some undocumented reason, on Windows we used the parent directory
    # lk: for start menu items. Mirror it here for backwards compatibility
    "start_menu": applications_dir().parent if platform.system() == "Windows" else applications_dir(),
}

def desktop_link_types() -> List:
    return list(__link_type.keys())
# fmt: on


def desktop_link_path(link_name: str, link_type: str) -> Path:
    """
    Creates the path of a shortcut link

    :param link_name:
        Name of the shortcut file
    :param link_type:
        "desktop" or "start_menu"
    :return Path:
        shortcut path
    """
    return __link_type[link_type].joinpath(f"{link_name}.{__link_suffix[platform.system()]['link']}")


def get_rare_executable() -> List[str]:
    logger.debug(f"Trying to find executable: {sys.executable}, {sys.argv}")
    # lk: detect if nuitka
    if "__compiled__" in globals():
        executable = [sys.executable]
    elif sys.argv[0].endswith("__main__.py"):
        executable = [sys.executable, "-m", "rare"]
    elif platform.system() == "Linux" or platform.system() == "Darwin" or platform.system() == "FreeBSD":
        if p := os.environ.get("APPIMAGE"):
            executable = [p]
        else:
            if sys.executable == os.path.abspath(sys.argv[0]):
                executable = [sys.executable]
            else:
                executable = [os.path.abspath(sys.argv[0])]
    elif platform.system() == "Windows":
        executable = [sys.executable]

        if sys.executable != os.path.abspath(sys.argv[0]):
            executable.append(os.path.abspath(sys.argv[0]))

        if executable[0].endswith("python.exe"):
            # be sure to start console-less then
            executable[0] = executable[0].replace("python.exe", "pythonw.exe")

        if executable[0].endswith("pythonw.exe"):
            if executable[1].endswith("rare"):
                executable[1] = executable[1] + ".exe"
    else:
        executable = [sys.executable]

    executable[0] = os.path.abspath(executable[0])
    return executable


def create_desktop_link(app_name: str, app_title: str = "", link_name: str = "", link_type="desktop") -> bool:
    """
    Creates a desktop or start menu shortcut link

    :param app_name:
        app_name or "rare_shortcut" for Rare itself
    :param app_title:
        the title shown in the shortcut
        (overrides to "Rare" for "rare_shortcut")
    :param link_name:
        the sanitized filename of the shortcut (use the folder_name attribute of RareGame)
        (overrides to "Rare" for "rare_shortcut")
    :param link_type:
        "desktop" or "start_menu"
    :return bool:
        True if successful else False
    """
    # macOS is based on Darwin
    if not desktop_links_supported():
        logger.error(f"Shortcut creation is not available on {platform.system()}")
        return False

    if link_type not in ["desktop", "start_menu"]:
        logger.error(f"Invalid link type {link_type}")
        return False

    for_rare = app_name == "rare_shortcut"

    if for_rare:
        icon_path = resources_path.joinpath("images", f"Rare.{desktop_icon_suffix()}")
        app_title = "Rare"
        link_name = "Rare"
    else:
        icon_path = image_dir().joinpath(app_name, f"icon.{desktop_icon_suffix()}")
        if not app_title or not link_name:
            logger.error("Missing app_title or link_name")
            return False

    shortcut_path = desktop_link_path(link_name, link_type)
    if not shortcut_path.parent.exists():
        logger.error(f"Parent directory {shortcut_path.parent} does not exist")
        return False
    else:
        logger.info(f"Creating shortcut for {app_title} at {shortcut_path}")

    if platform.system() == "Linux" or platform.system() == "FreeBSD":
        executable = get_rare_executable()
        executable = shlex.join(executable)
        if not for_rare:
            executable = f"{executable} launch {app_name}"

        with shortcut_path.open(mode="w") as desktop_file:
            desktop_file.write(
                "[Desktop Entry]\n"
                f"Name={app_title}\n"
                "Type=Application\n"
                "Categories=Game;\n"
                f"Icon={icon_path}\n"
                f"Exec={executable}\n"
                "Terminal=false\n"
                "StartupWMClass=Rare\n"
            )
        # shortcut_path.chmod(0o755)
        return True

    elif platform.system() == "Windows":
        # Add shortcut
        shell = Dispatch("WScript.Shell")
        shortcut = shell.CreateShortCut(str(shortcut_path))

        executable = get_rare_executable()
        arguments = []

        if len(executable) > 1:
            arguments.extend(executable[1:])
        executable = executable[0]

        if not for_rare:
            arguments.extend(["launch", app_name])
        shortcut.Targetpath = executable
        # Maybe there is a better solution, but windows does not accept single quotes (Windows is weird)
        shortcut.Arguments = shlex.join(arguments).replace("'", '"')

        shortcut.Description = app_title
        if for_rare:
            shortcut.WorkingDirectory = str(home_dir())

        # Icon
        shortcut.IconLocation = str(icon_path.absolute())

        shortcut.save()
        return True



from typing import Union

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFontMetrics, QResizeEvent
from PyQt5.QtWidgets import QLabel, QWIDGETSIZE_MAX


class ElideLabel(QLabel):

    def __init__(self, text="", parent=None):
        super(ElideLabel, self).__init__(parent=parent)
        self.__text = text
        self.__fm = QFontMetrics(self.font())
        self.__tooltip = ""
        self.setFixedHeight(True)
        self.setWordWrap(True)
        self.setText(text)

    def setText(self, a0: str) -> None:
        self.__text = a0
        self.__setElideText(a0)

    def __setElideText(self, a0: str):
        elided_text = self.__fm.elidedText(
            a0, Qt.ElideRight,
            self.width() - (self.contentsMargins().left() + self.contentsMargins().right())
        )
        if not self.__tooltip:
            if self.__fm.boundingRect(elided_text).width() < self.__fm.boundingRect(self.__text).width():
                super(ElideLabel, self).setToolTip(self.__text)
            else:
                super(ElideLabel, self).setToolTip("")
        super(ElideLabel, self).setText(elided_text)

    def setToolTip(self, a0: str) -> None:
        self.__tooltip = a0
        super(ElideLabel, self).setToolTip(a0)

    def resizeEvent(self, a0: QResizeEvent) -> None:
        self.__setElideText(self.__text)
        super(ElideLabel, self).resizeEvent(a0)

    def setFixedHeight(self, h: Union[int, bool]) -> None:
        if isinstance(h, bool):
            super(ElideLabel, self).setFixedHeight(self.__fm.height() if h else QWIDGETSIZE_MAX)
        elif isinstance(h, int):
            super(ElideLabel, self).setFixedHeight(h)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/dialogs/sync_save_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_SyncSaveDialog(object):
    def setupUi(self, SyncSaveDialog):
        SyncSaveDialog.setObjectName("SyncSaveDialog")
        SyncSaveDialog.resize(648, 394)
        self.verticalLayout = QtWidgets.QVBoxLayout(SyncSaveDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.title_label = QtWidgets.QLabel(SyncSaveDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.title_label.sizePolicy().hasHeightForWidth())
        self.title_label.setSizePolicy(sizePolicy)
        self.title_label.setObjectName("title_label")
        self.verticalLayout.addWidget(self.title_label)
        self.sync_widget_layout = QtWidgets.QHBoxLayout()
        self.sync_widget_layout.setObjectName("sync_widget_layout")
        self.verticalLayout.addLayout(self.sync_widget_layout)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.cancel_button = QtWidgets.QPushButton(SyncSaveDialog)
        self.cancel_button.setObjectName("cancel_button")
        self.horizontalLayout_2.addWidget(self.cancel_button)
        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.retranslateUi(SyncSaveDialog)

    def retranslateUi(self, SyncSaveDialog):
        _translate = QtCore.QCoreApplication.translate
        SyncSaveDialog.setWindowTitle(_translate("SyncSaveDialog", "Sync saves with cloud"))
        self.title_label.setText(_translate("SyncSaveDialog", "Select save, you want to use for "))
        self.cancel_button.setText(_translate("SyncSaveDialog", "Cancel"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    SyncSaveDialog = QtWidgets.QDialog()
    ui = Ui_SyncSaveDialog()
    ui.setupUi(SyncSaveDialog)
    SyncSaveDialog.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/games/import_sync/eos_widget.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_EosWidget(object):
    def setupUi(self, EosWidget):
        EosWidget.setObjectName("EosWidget")
        EosWidget.resize(586, 146)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(EosWidget.sizePolicy().hasHeightForWidth())
        EosWidget.setSizePolicy(sizePolicy)
        EosWidget.setWindowTitle("GroupBox")
        self.eos_layout = QtWidgets.QHBoxLayout(EosWidget)
        self.eos_layout.setSizeConstraint(QtWidgets.QLayout.SetDefaultConstraint)
        self.eos_layout.setObjectName("eos_layout")
        self.overlay_stack = QtWidgets.QStackedWidget(EosWidget)
        self.overlay_stack.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.overlay_stack.setFrameShadow(QtWidgets.QFrame.Raised)
        self.overlay_stack.setObjectName("overlay_stack")
        self.overlay_info_page = QtWidgets.QWidget()
        self.overlay_info_page.setObjectName("overlay_info_page")
        self.formLayout_3 = QtWidgets.QFormLayout(self.overlay_info_page)
        self.formLayout_3.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.formLayout_3.setFormAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.formLayout_3.setObjectName("formLayout_3")
        self.installed_version_info_lbl = QtWidgets.QLabel(self.overlay_info_page)
        self.installed_version_info_lbl.setObjectName("installed_version_info_lbl")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.installed_version_info_lbl)
        self.installed_version_lbl = QtWidgets.QLabel(self.overlay_info_page)
        self.installed_version_lbl.setText("error")
        self.installed_version_lbl.setObjectName("installed_version_lbl")
        self.formLayout_3.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.installed_version_lbl)
        self.installed_path_info_lbl = QtWidgets.QLabel(self.overlay_info_page)
        self.installed_path_info_lbl.setObjectName("installed_path_info_lbl")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.installed_path_info_lbl)
        self.installed_path_lbl = QtWidgets.QLabel(self.overlay_info_page)
        self.installed_path_lbl.setText("error")
        self.installed_path_lbl.setObjectName("installed_path_lbl")
        self.formLayout_3.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.installed_path_lbl)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.uninstall_button = QtWidgets.QPushButton(self.overlay_info_page)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.uninstall_button.sizePolicy().hasHeightForWidth())
        self.uninstall_button.setSizePolicy(sizePolicy)
        self.uninstall_button.setMaximumSize(QtCore.QSize(150, 16777215))
        self.uninstall_button.setObjectName("uninstall_button")
        self.horizontalLayout.addWidget(self.uninstall_button)
        self.update_check_button = QtWidgets.QPushButton(self.overlay_info_page)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.update_check_button.sizePolicy().hasHeightForWidth())
        self.update_check_button.setSizePolicy(sizePolicy)
        self.update_check_button.setMaximumSize(QtCore.QSize(150, 16777215))
        self.update_check_button.setObjectName("update_check_button")
        self.horizontalLayout.addWidget(self.update_check_button)
        self.update_button = QtWidgets.QPushButton(self.overlay_info_page)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.update_button.sizePolicy().hasHeightForWidth())
        self.update_button.setSizePolicy(sizePolicy)
        self.update_button.setMaximumSize(QtCore.QSize(150, 16777215))
        self.update_button.setObjectName("update_button")
        self.horizontalLayout.addWidget(self.update_button)
        self.formLayout_3.setLayout(3, QtWidgets.QFormLayout.SpanningRole, self.horizontalLayout)
        spacerItem1 = QtWidgets.QSpacerItem(6, 6, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout_3.setItem(2, QtWidgets.QFormLayout.SpanningRole, spacerItem1)
        self.overlay_stack.addWidget(self.overlay_info_page)
        self.overlay_install_page = QtWidgets.QWidget()
        self.overlay_install_page.setObjectName("overlay_install_page")
        self.formLayout = QtWidgets.QFormLayout(self.overlay_install_page)
        self.formLayout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.formLayout.setFormAlignment(QtCore.Qt.AlignBottom|QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(self.overlay_install_page)
        self.label.setObjectName("label")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.SpanningRole, self.label)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.install_button = QtWidgets.QPushButton(self.overlay_install_page)
        self.install_button.setObjectName("install_button")
        self.horizontalLayout_3.addWidget(self.install_button)
        self.formLayout.setLayout(2, QtWidgets.QFormLayout.SpanningRole, self.horizontalLayout_3)
        spacerItem3 = QtWidgets.QSpacerItem(6, 6, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(1, QtWidgets.QFormLayout.SpanningRole, spacerItem3)
        self.overlay_stack.addWidget(self.overlay_install_page)
        self.eos_layout.addWidget(self.overlay_stack)
        self.enable_frame = QtWidgets.QFrame(EosWidget)
        self.enable_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.enable_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.enable_frame.setObjectName("enable_frame")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.enable_frame)
        self.verticalLayout.setObjectName("verticalLayout")
        self.select_pfx_combo = QtWidgets.QComboBox(self.enable_frame)
        self.select_pfx_combo.setObjectName("select_pfx_combo")
        self.verticalLayout.addWidget(self.select_pfx_combo)
        self.enabled_cb = QtWidgets.QCheckBox(self.enable_frame)
        self.enabled_cb.setObjectName("enabled_cb")
        self.verticalLayout.addWidget(self.enabled_cb)
        self.enabled_info_label = QtWidgets.QLabel(self.enable_frame)
        font = QtGui.QFont()
        font.setItalic(True)
        self.enabled_info_label.setFont(font)
        self.enabled_info_label.setText("")
        self.enabled_info_label.setObjectName("enabled_info_label")
        self.verticalLayout.addWidget(self.enabled_info_label)
        self.eos_layout.addWidget(self.enable_frame)

        self.retranslateUi(EosWidget)
        self.overlay_stack.setCurrentIndex(0)

    def retranslateUi(self, EosWidget):
        _translate = QtCore.QCoreApplication.translate
        EosWidget.setTitle(_translate("EosWidget", "Epic Overlay"))
        self.installed_version_info_lbl.setText(_translate("EosWidget", "Version"))
        self.installed_path_info_lbl.setText(_translate("EosWidget", "Location"))
        self.uninstall_button.setText(_translate("EosWidget", "Uninstall"))
        self.update_check_button.setText(_translate("EosWidget", "Check for update"))
        self.update_button.setText(_translate("EosWidget", "Update"))
        self.label.setText(_translate("EosWidget", "Epic Overlay Services is not installed"))
        self.install_button.setText(_translate("EosWidget", "Install"))
        self.enabled_cb.setText(_translate("EosWidget", "Activated"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    EosWidget = QtWidgets.QGroupBox()
    ui = Ui_EosWidget()
    ui.setupUi(EosWidget)
    EosWidget.show()
    sys.exit(app.exec_())



import json
import platform
import subprocess
import time
import traceback
from argparse import Namespace
from logging import getLogger
from signal import signal, SIGINT, SIGTERM, strsignal
from typing import Optional

from PyQt5.QtCore import QObject, QProcess, pyqtSignal, QUrl, QRunnable, QThreadPool, QSettings, Qt, pyqtSlot
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtNetwork import QLocalServer, QLocalSocket
from PyQt5.QtWidgets import QApplication
from legendary.models.game import SaveGameStatus

from rare.components.dialogs.cloud_save_dialog import CloudSaveDialog
from rare.lgndr.core import LegendaryCore
from rare.models.base_game import RareGameSlim
from rare.models.launcher import ErrorModel, Actions, FinishedModel, BaseModel, StateChangedModel
from rare.widgets.rare_app import RareApp, RareAppException
from .console import Console
from .lgd_helper import get_launch_args, InitArgs, get_configured_process, LaunchArgs, GameArgsError

DETACHED_APP_NAMES = [
    "0a2d9f6403244d12969e11da6713137b"  # Fall Guys
    "Fortnite"
]


class PreLaunchThread(QRunnable):
    class Signals(QObject):
        ready_to_launch = pyqtSignal(LaunchArgs)
        started_pre_launch_command = pyqtSignal()
        pre_launch_command_finished = pyqtSignal(int)  # exit_code
        error_occurred = pyqtSignal(str)

    def __init__(self, core: LegendaryCore, args: InitArgs, rgame: RareGameSlim, sync_action=None):
        super(PreLaunchThread, self).__init__()
        self.logger = getLogger(type(self).__name__)
        self.core = core
        self.signals = self.Signals()
        self.args = args
        self.rgame = rgame
        self.sync_action = sync_action

    def run(self) -> None:
        self.logger.info(f"Sync action: {self.sync_action}")
        if self.sync_action == CloudSaveDialog.UPLOAD:
            self.rgame.upload_saves(False)
        elif self.sync_action == CloudSaveDialog.DOWNLOAD:
            self.rgame.download_saves(False)
        else:
            self.logger.info("No sync action")

        args = self.prepare_launch(self.args)
        if not args:
            return
        self.signals.ready_to_launch.emit(args)

    def prepare_launch(self, args: InitArgs) -> Optional[LaunchArgs]:
        try:
            launch_args = get_launch_args(self.core, args)
        except Exception as e:
            self.signals.error_occurred.emit(str(e))
            return None
        if not launch_args:
            return None

        if launch_args.pre_launch_command:
            proc = get_configured_process()
            proc.setProcessEnvironment(launch_args.environment)
            self.signals.started_pre_launch_command.emit()
            proc.start(launch_args.pre_launch_command[0], launch_args.pre_launch_command[1:])
            if launch_args.pre_launch_wait:
                proc.waitForFinished(-1)
        return launch_args


class SyncCheckWorker(QRunnable):
    class Signals(QObject):
        sync_state_ready = pyqtSignal()
        error_occurred = pyqtSignal(str)

    def __init__(self, core: LegendaryCore, rgame: RareGameSlim):
        super().__init__()
        self.signals = self.Signals()
        self.core = core
        self.rgame = rgame

    def run(self) -> None:
        try:
            self.rgame.update_saves()
        except Exception as e:
            self.signals.error_occurred.emit(str(e))
            return
        self.signals.sync_state_ready.emit()


class RareLauncherException(RareAppException):
    def __init__(self, app: 'RareLauncher', args: Namespace, parent=None):
        super(RareLauncherException, self).__init__(parent=parent)
        self.__app = app
        self.__args = args

    def _handler(self, exc_type, exc_value, exc_tb) -> bool:
        try:
            self.__app.send_message(ErrorModel(
                app_name=self.__args.app_name,
                action=Actions.error,
                error_string="".join(traceback.format_exception(exc_type, exc_value, exc_tb))
            ))
        except RuntimeError:
            pass
        return False


class RareLauncher(RareApp):
    exit_app = pyqtSignal()

    def __init__(self, args: InitArgs):
        super(RareLauncher, self).__init__(args, f"{type(self).__name__}_{args.app_name}_{{0}}.log")
        self.socket: Optional[QLocalSocket] = None
        self.console: Optional[Console] = None
        self.game_process: QProcess = QProcess(self)
        self.server: QLocalServer = QLocalServer(self)

        self._hook.deleteLater()
        self._hook = RareLauncherException(self, args, self)

        self.success: bool = False
        self.no_sync_on_exit = False
        self.args = args
        self.core = LegendaryCore()
        game = self.core.get_game(args.app_name)
        if not game:
            self.logger.error(f"Game {args.app_name} not found. Exiting")
            return
        self.rgame = RareGameSlim(self.core, game)

        lang = self.settings.value("language", self.core.language_code, type=str)
        self.load_translator(lang)

        if QSettings().value("show_console", False, bool):
            self.console = Console()
            self.console.show()

        self.game_process.finished.connect(self.__process_finished)
        self.game_process.errorOccurred.connect(self.__process_errored)
        if self.console:
            self.game_process.readyReadStandardOutput.connect(self.__proc_log_stdout)
            self.game_process.readyReadStandardError.connect(self.__proc_log_stderr)
            self.console.term.connect(self.__proc_term)
            self.console.kill.connect(self.__proc_kill)

        ret = self.server.listen(f"rare_{args.app_name}")
        if not ret:
            self.logger.error(self.server.errorString())
            self.logger.info("Server is running")
            self.server.close()
            return
        self.server.newConnection.connect(self.new_server_connection)

        self.success = True
        self.start_time = time.time()

    @pyqtSlot()
    def __proc_log_stdout(self):
        self.console.log(
            self.game_process.readAllStandardOutput().data().decode("utf-8", "ignore")
        )

    @pyqtSlot()
    def __proc_log_stderr(self):
        self.console.error(
            self.game_process.readAllStandardError().data().decode("utf-8", "ignore")
        )

    @pyqtSlot()
    def __proc_term(self):
        self.game_process.terminate()

    @pyqtSlot()
    def __proc_kill(self):
        self.game_process.kill()

    def new_server_connection(self):
        if self.socket is not None:
            try:
                self.socket.disconnectFromServer()
            except RuntimeError:
                pass
        self.logger.info("New connection")
        self.socket = self.server.nextPendingConnection()
        self.socket.disconnected.connect(self.socket_disconnected)
        self.socket.flush()

    def socket_disconnected(self):
        self.logger.info("Server disconnected")
        self.socket.deleteLater()
        self.socket = None

    def send_message(self, message: BaseModel):
        if self.socket:
            self.socket.write(json.dumps(vars(message)).encode("utf-8"))
            self.socket.flush()
        else:
            self.logger.error("Can't send message")

    def check_saves_finished(self, exit_code: int):
        self.rgame.signals.widget.update.connect(lambda: self.on_exit(exit_code))

        state, (dt_local, dt_remote) = self.rgame.save_game_state

        if state == SaveGameStatus.LOCAL_NEWER and not self.no_sync_on_exit:
            action = CloudSaveDialog.UPLOAD
        else:
            action = CloudSaveDialog(self.rgame.igame, dt_local, dt_remote).get_action()

        if action == CloudSaveDialog.UPLOAD:
            if self.console:
                self.console.log("Uploading saves...")
            self.rgame.upload_saves()
        elif action == CloudSaveDialog.DOWNLOAD:
            if self.console:
                self.console.log("Downloading saves...")
            self.rgame.download_saves()
        else:
            self.on_exit(exit_code)

    @pyqtSlot(int, QProcess.ExitStatus)
    def __process_finished(self, exit_code: int, exit_status: QProcess.ExitStatus):
        self.logger.info("Game finished")

        if self.rgame.auto_sync_saves:
            self.check_saves_finished(exit_code)
        else:
            self.on_exit(exit_code)

    @pyqtSlot(QProcess.ProcessError)
    def __process_errored(self, error: QProcess.ProcessError):
        self.error_occurred(self.game_process.errorString())

    def on_exit(self, exit_code: int):
        if self.console:
            self.console.on_process_exit(self.core.get_game(self.rgame.app_name).app_title, exit_code)

        self.send_message(
            FinishedModel(
                action=Actions.finished,
                app_name=self.rgame.app_name,
                exit_code=exit_code,
                playtime=int(time.time() - self.start_time)
            )

        )
        self.stop()

    @pyqtSlot(object)
    def launch_game(self, args: LaunchArgs):
        # should never happen
        if not args:
            self.stop()
            return
        if self.console:
            self.console.set_env(args.environment)
        self.start_time = time.time()

        if args.is_origin_game:
            QDesktopServices.openUrl(QUrl(args.executable))
            self.stop()  # stop because it is no subprocess
            return

        if args.working_directory:
            self.game_process.setWorkingDirectory(args.working_directory)
        self.game_process.setProcessEnvironment(args.environment)
        # send start message after process started
        self.game_process.started.connect(lambda: self.send_message(
            StateChangedModel(
                action=Actions.state_update, app_name=self.rgame.app_name,
                new_state=StateChangedModel.States.started
            )
        ))
        if self.rgame.app_name in DETACHED_APP_NAMES and platform.system() == "Windows":
            self.game_process.deleteLater()
            subprocess.Popen([args.executable] + args.arguments, cwd=args.working_directory,
                             env={i: args.environment.value(i) for i in args.environment.keys()})
            if self.console:
                self.console.log("Launching game as a detached process")
            self.stop()
            return
        if self.args.dry_run:
            self.logger.info("Dry run activated")
            if self.console:
                self.console.log(f"{args.executable} {' '.join(args.arguments)}")
                self.console.log(f"Do not start {self.rgame.app_name}")
                self.console.accept_close = True
            print(args.executable, " ".join(args.arguments))
            self.stop()
            return
        self.game_process.start(args.executable, args.arguments)

    def error_occurred(self, error_str: str):
        self.logger.warning(error_str)
        if self.console:
            self.console.on_process_exit(self.core.get_game(self.rgame.app_name).app_title, error_str)
        self.send_message(ErrorModel(
            error_string=error_str, app_name=self.rgame.app_name,
            action=Actions.error)
        )
        self.stop()

    def start_prepare(self, sync_action=None):
        worker = PreLaunchThread(self.core, self.args, self.rgame, sync_action)
        worker.signals.ready_to_launch.connect(self.launch_game)
        worker.signals.error_occurred.connect(self.error_occurred)
        # worker.signals.started_pre_launch_command(None)

        QThreadPool.globalInstance().start(worker)

    def sync_ready(self):
        if self.rgame.is_save_up_to_date:
            if self.console:
                self.console.log("Sync worker ready. Sync not required")
            self.start_prepare()
            return

        _, (dt_local, dt_remote) = self.rgame.save_game_state
        dlg = CloudSaveDialog(self.rgame.igame, dt_local, dt_remote)
        action = dlg.get_action()
        if action == CloudSaveDialog.CANCEL:
            self.no_sync_on_exit = True
        if self.console:
            if action == CloudSaveDialog.DOWNLOAD:
                self.console.log("Downloading saves...")
            elif action == CloudSaveDialog.UPLOAD:
                self.console.log("Uploading saves...")
        self.start_prepare(action)

    def start(self, args: InitArgs):
        if not args.offline:
            try:
                if not self.core.login():
                    raise ValueError("You are not logged in")
            except ValueError:
                # automatically launch offline if available
                self.logger.error("Not logged in. Trying to launch the game in offline mode")
                args.offline = True

        if not args.offline and self.rgame.auto_sync_saves:
            self.logger.info("Start sync worker")
            worker = SyncCheckWorker(self.core, self.rgame)
            worker.signals.error_occurred.connect(self.error_occurred)
            worker.signals.sync_state_ready.connect(self.sync_ready)
            QThreadPool.globalInstance().start(worker)
            return
        else:
            self.start_prepare()

    def stop(self):
        try:
            if self.console:
                self.game_process.readyReadStandardOutput.disconnect()
                self.game_process.readyReadStandardError.disconnect()
            if self.game_process.receivers(self.game_process.finished):
                self.game_process.finished.disconnect()
            if self.game_process.receivers(self.game_process.errorOccurred):
                self.game_process.errorOccurred.disconnect()
        except TypeError as e:
            self.logger.error(f"Failed to disconnect process signals: {e}")

        self.logger.info("Stopping server")
        try:
            self.server.close()
            self.server.deleteLater()
        except RuntimeError:
            pass
        self.processEvents()
        if not self.console:
            self.exit()
        else:
            self.console.on_process_exit(self.rgame.app_name, 0)


def launch(args: Namespace) -> int:
    args = InitArgs.from_argparse(args)

    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

    app = RareLauncher(args)
    app.setQuitOnLastWindowClosed(True)

    # This prevents ghost QLocalSockets, which block the name, which makes it unable to start
    # No handling for SIGKILL
    def sighandler(s, frame):
        app.logger.info(f"{strsignal(s)} received. Stopping")
        app.stop()
        app.exit(1)
    signal(SIGINT, sighandler)
    signal(SIGTERM, sighandler)

    if not app.success:
        app.stop()
        app.exit(1)
        return 1
    app.start(args)
    # app.exit_app.connect(lambda: app.exit(0))

    return app.exec_()



from logging import getLogger

from PyQt5.QtCore import QObject, pyqtSignal
from legendary.core import LegendaryCore

from rare.lgndr.cli import LegendaryCLI
from rare.lgndr.glue.arguments import LgndrUninstallGameArgs
from rare.lgndr.glue.monkeys import LgndrIndirectStatus
from rare.models.game import RareGame
from rare.models.install import UninstallOptionsModel
from rare.utils import config_helper
from rare.utils.paths import desktop_links_supported, desktop_link_types, desktop_link_path
from .worker import Worker

logger = getLogger("UninstallWorker")


# TODO: You can use RareGame directly here once this is called inside RareCore and skip metadata fetch
def uninstall_game(core: LegendaryCore, app_name: str, keep_files=False, keep_config=False):
    game = core.get_game(app_name)

    # remove shortcuts link
    if desktop_links_supported():
        for link_type in desktop_link_types():
            link_path = desktop_link_path(
                game.metadata.get("customAttributes", {}).get("FolderName", {}).get("value"), link_type
            )
            if link_path.exists():
                link_path.unlink(missing_ok=True)

    status = LgndrIndirectStatus()
    LegendaryCLI(core).uninstall_game(
        LgndrUninstallGameArgs(
            app_name=app_name,
            keep_files=keep_files,
            skip_uninstaller=False,
            yes=True,
            indirect_status=status,
        )
    )
    if not keep_config:
        logger.info("Removing sections in config file")
        config_helper.remove_section(app_name)
        config_helper.remove_section(f"{app_name}.env")

        config_helper.save_config()

    return status.success, status.message


class UninstallWorker(Worker):
    class Signals(QObject):
        result = pyqtSignal(RareGame, bool, str)

    def __init__(self, core: LegendaryCore, rgame: RareGame, options: UninstallOptionsModel):
        super(UninstallWorker, self).__init__()
        self.signals = UninstallWorker.Signals()
        self.core = core
        self.rgame = rgame
        self.options = options

    def run_real(self) -> None:
        self.rgame.state = RareGame.State.UNINSTALLING
        success, message = uninstall_game(
            self.core, self.rgame.app_name, self.options.keep_files, self.options.keep_config
        )
        self.rgame.state = RareGame.State.IDLE
        self.signals.result.emit(self.rgame, success, message)



from PyQt5.QtCore import QObject


# Class to use QObject.tr()
class Constants(QObject):
    def __init__(self):
        super(Constants, self).__init__()
        self.categories = sorted(
            [
                (self.tr("Action"), "1216"),
                (self.tr("Adventure"), "1117"),
                (self.tr("Puzzle"), "1298"),
                (self.tr("Open world"), "1307"),
                (self.tr("Racing"), "1212"),
                (self.tr("RPG"), "1367"),
                (self.tr("Shooter"), "1210"),
                (self.tr("Strategy"), "1115"),
                (self.tr("Survival"), "1080"),
                (self.tr("First Person"), "1294"),
                (self.tr("Indie"), "1263"),
                (self.tr("Simulation"), "1393"),
                (self.tr("Sport"), "1283"),
            ],
            key=lambda x: x[0],
        )

        self.platforms = [
            ("MacOS", "9548"),
            ("Windows", "9547"),
        ]
        self.others = [
            (self.tr("Single player"), "1370"),
            (self.tr("Multiplayer"), "1203"),
            (self.tr("Controller"), "9549"),
            (self.tr("Co-op"), "1264"),
        ]

        self.types = [
            (self.tr("Editor"), "editors"),
            (self.tr("Game"), "games/edition/base"),
            (self.tr("Bundle"), "bundles/games"),
            (self.tr("Add-on"), "addons"),
            (self.tr("Apps"), "software/edition/base"),
        ]


game_query = (
    "query searchStoreQuery($allowCountries: String, $category: String, $count: Int, $country: String!, "
    "$keywords: String, $locale: String, $namespace: String, $withMapping: Boolean = false, $itemNs: String, "
    "$sortBy: String, $sortDir: String, $start: Int, $tag: String, $releaseDate: String, $withPrice: Boolean "
    "= false, $withPromotions: Boolean = false, $priceRange: String, $freeGame: Boolean, $onSale: Boolean, "
    "$effectiveDate: String) {\n  Catalog {\n    searchStore(\n      allowCountries: $allowCountries\n      "
    "category: $category\n      count: $count\n      country: $country\n      keywords: $keywords\n      "
    "locale: $locale\n      namespace: $namespace\n      itemNs: $itemNs\n      sortBy: $sortBy\n      "
    "sortDir: $sortDir\n      releaseDate: $releaseDate\n      start: $start\n      tag: $tag\n      "
    "priceRange: $priceRange\n      freeGame: $freeGame\n      onSale: $onSale\n      effectiveDate: "
    "$effectiveDate\n    ) {\n      elements {\n        title\n        id\n        namespace\n        "
    "description\n        effectiveDate\n        keyImages {\n          type\n          url\n        }\n     "
    "   currentPrice\n        seller {\n          id\n          name\n        }\n        productSlug\n       "
    " urlSlug\n        url\n        tags {\n          id\n        }\n        items {\n          id\n         "
    " namespace\n        }\n        customAttributes {\n          key\n          value\n        }\n        "
    "categories {\n          path\n        }\n        catalogNs @include(if: $withMapping) {\n          "
    'mappings(pageType: "productHome") {\n            pageSlug\n            pageType\n          }\n        '
    "}\n        offerMappings @include(if: $withMapping) {\n          pageSlug\n          pageType\n        "
    "}\n        price(country: $country) @include(if: $withPrice) {\n          totalPrice {\n            "
    "discountPrice\n            originalPrice\n            voucherDiscount\n            discount\n           "
    " currencyCode\n            currencyInfo {\n              decimals\n            }\n            fmtPrice("
    "locale: $locale) {\n              originalPrice\n              discountPrice\n              "
    "intermediatePrice\n            }\n          }\n          lineOffers {\n            appliedRules {\n     "
    "         id\n              endDate\n              discountSetting {\n                discountType\n     "
    "         }\n            }\n          }\n        }\n        promotions(category: $category) @include(if: "
    "$withPromotions) {\n          promotionalOffers {\n            promotionalOffers {\n              "
    "startDate\n              endDate\n              discountSetting {\n                discountType\n       "
    "         discountPercentage\n              }\n            }\n          }\n          "
    "upcomingPromotionalOffers {\n            promotionalOffers {\n              startDate\n              "
    "endDate\n              discountSetting {\n                discountType\n                "
    "discountPercentage\n              }\n            }\n          }\n        }\n      }\n      paging {\n   "
    "     count\n        total\n      }\n    }\n  }\n}\n "
)

search_query = (
    "query searchStoreQuery($allowCountries: String, $category: String, $count: Int, $country: String!, "
    "$keywords: String, $locale: String, $namespace: String, $withMapping: Boolean = false, $itemNs: String, "
    "$sortBy: String, $sortDir: String, $start: Int, $tag: String, $releaseDate: String, $withPrice: Boolean = "
    "false, $withPromotions: Boolean = false, $priceRange: String, $freeGame: Boolean, $onSale: Boolean, "
    "$effectiveDate: String) {\n  Catalog {\n    searchStore(\n      allowCountries: $allowCountries\n      "
    "category: $category\n      count: $count\n      country: $country\n      keywords: $keywords\n      locale: "
    "$locale\n      namespace: $namespace\n      itemNs: $itemNs\n      sortBy: $sortBy\n      sortDir: "
    "$sortDir\n      releaseDate: $releaseDate\n      start: $start\n      tag: $tag\n      priceRange: "
    "$priceRange\n      freeGame: $freeGame\n      onSale: $onSale\n      effectiveDate: $effectiveDate\n    ) {"
    "\n      elements {\n        title\n        id\n        namespace\n        description\n        "
    "effectiveDate\n        keyImages {\n          type\n          url\n        }\n        currentPrice\n        "
    "seller {\n          id\n          name\n        }\n        productSlug\n        urlSlug\n        url\n       "
    " tags {\n          id\n        }\n        items {\n          id\n          namespace\n        }\n        "
    "customAttributes {\n          key\n          value\n        }\n        categories {\n          path\n        "
    '}\n        catalogNs @include(if: $withMapping) {\n          mappings(pageType: "productHome") {\n         '
    "   pageSlug\n            pageType\n          }\n        }\n        offerMappings @include(if: $withMapping) "
    "{\n          pageSlug\n          pageType\n        }\n        price(country: $country) @include(if: "
    "$withPrice) {\n          totalPrice {\n            discountPrice\n            originalPrice\n            "
    "voucherDiscount\n            discount\n            currencyCode\n            currencyInfo {\n              "
    "decimals\n            }\n            fmtPrice(locale: $locale) {\n              originalPrice\n              "
    "discountPrice\n              intermediatePrice\n            }\n          }\n          lineOffers {\n         "
    "   appliedRules {\n              id\n              endDate\n              discountSetting {\n                "
    "discountType\n              }\n            }\n          }\n        }\n        promotions(category: "
    "$category) @include(if: $withPromotions) {\n          promotionalOffers {\n            promotionalOffers {\n "
    "             startDate\n              endDate\n              discountSetting {\n                "
    "discountType\n                discountPercentage\n              }\n            }\n          }\n          "
    "upcomingPromotionalOffers {\n            promotionalOffers {\n              startDate\n              "
    "endDate\n              discountSetting {\n                discountType\n                discountPercentage\n "
    "             }\n            }\n          }\n        }\n      }\n      paging {\n        count\n        "
    "total\n      }\n    }\n  }\n}\n "
)

wishlist_query = '\n    query wishlistQuery($country:String!, $locale:String) {\n        Wishlist {\n            wishlistItems {\n                elements {\n                    id\n                    order\n                    created\n                    offerId\n                    updated\n                    namespace\n                    \n    offer {\n        productSlug\n        urlSlug\n        title\n        id\n        namespace\n        offerType\n        expiryDate\n        status\n        isCodeRedemptionOnly\n        description\n        effectiveDate\n        keyImages {\n            type\n            url\n        }\n        seller {\n            id\n            name\n        }\n        productSlug\n        urlSlug\n        items {\n            id\n            namespace\n        }\n        customAttributes {\n            key\n            value\n        }\n        catalogNs {\n            mappings(pageType: "productHome") {\n                pageSlug\n                pageType\n            }\n        }\n        offerMappings {\n            pageSlug\n            pageType\n        }\n        categories {\n            path\n        }\n        price(country: $country) {\n            totalPrice {\n                discountPrice\n                originalPrice\n                voucherDiscount\n                discount\n                fmtPrice(locale: $locale) {\n                    originalPrice\n                    discountPrice\n                    intermediatePrice\n                }\n                currencyCode\n                currencyInfo {\n                    decimals\n                    symbol\n                }\n            }\n            lineOffers {\n                appliedRules {\n                    id\n                    endDate\n                }\n            }\n        }\n    }\n\n                }\n            }\n        }\n    }\n'
add_to_wishlist_query = "\n    mutation removeFromWishlistMutation($namespace: String!, $offerId: String!, $operation: RemoveOperation!) {\n        Wishlist {\n            removeFromWishlist(namespace: $namespace, offerId: $offerId, operation: $operation) {\n                success\n            }\n        }\n    }\n"
remove_from_wishlist_query = "\n    mutation removeFromWishlistMutation($namespace: String!, $offerId: String!, $operation: RemoveOperation!) {\n        Wishlist {\n            removeFromWishlist(namespace: $namespace, offerId: $offerId, operation: $operation) {\n                success\n            }\n        }\n    }\n"
coupon_query = "\n    query getCoupons($currencyCountry: String!, $identityId: String!, $locale: String) {\n        CodeRedemption {\n            coupons(currencyCountry: $currencyCountry, identityId: $identityId, includeSalesEventInfo: true) {\n                code\n                codeStatus\n                codeType\n                consumptionMetadata {\n                    amountDisplay {\n                      amount\n                      currency\n                      placement\n                      symbol\n                    }\n                    minSalesPriceDisplay {\n                      amount\n                      currency\n                      placement\n                      symbol\n                    }\n                }\n                endDate\n                namespace\n                salesEvent(locale: $locale) {\n                    eventName\n                    eventSlug\n                    voucherImages {\n                      type\n                      url\n                    }\n                    voucherLink\n                }\n                startDate\n            }\n        }\n    }\n"



from rare.components.tabs.settings.widgets.linux import LinuxSettings
from rare.shared import ArgumentsSingleton
from rare.widgets.side_tab import SideTabWidget
from .about import About
from .debug import DebugSettings
from .game_settings import DefaultGameSettings
from .legendary import LegendarySettings
from .rare import RareSettings


class SettingsTab(SideTabWidget):
    def __init__(self, parent=None):
        super(SettingsTab, self).__init__(parent=parent)
        self.args = ArgumentsSingleton()

        self.rare_index = self.addTab(RareSettings(self), "Rare")
        self.legendary_index = self.addTab(LegendarySettings(self), "Legendary")
        self.settings_index = self.addTab(DefaultGameSettings(True, self), self.tr("Default Settings"))

        self.about = About(self)
        self.about_index = self.addTab(self.about, "About", "About")
        self.about.update_available_ready.connect(
            lambda: self.tabBar().setTabText(self.about_index, "About (!)")
        )

        if self.args.debug:
            self.debug_index = self.addTab(DebugSettings(self), "Debug")

        self.setCurrentIndex(self.rare_index)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/dialogs/install_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_InstallDialog(object):
    def setupUi(self, InstallDialog):
        InstallDialog.setObjectName("InstallDialog")
        InstallDialog.resize(272, 238)
        InstallDialog.setWindowTitle("InstallDialog")
        self.install_dialog_layout = QtWidgets.QFormLayout(InstallDialog)
        self.install_dialog_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.install_dialog_layout.setObjectName("install_dialog_layout")
        self.install_dialog_label = QtWidgets.QLabel(InstallDialog)
        self.install_dialog_label.setObjectName("install_dialog_label")
        self.install_dialog_layout.setWidget(0, QtWidgets.QFormLayout.SpanningRole, self.install_dialog_label)
        self.install_dir_label = QtWidgets.QLabel(InstallDialog)
        self.install_dir_label.setObjectName("install_dir_label")
        self.install_dialog_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.install_dir_label)
        self.platform_label = QtWidgets.QLabel(InstallDialog)
        self.platform_label.setObjectName("platform_label")
        self.install_dialog_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.platform_label)
        self.platform_combo = QtWidgets.QComboBox(InstallDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.platform_combo.sizePolicy().hasHeightForWidth())
        self.platform_combo.setSizePolicy(sizePolicy)
        self.platform_combo.setObjectName("platform_combo")
        self.install_dialog_layout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.platform_combo)
        self.shortcut_label = QtWidgets.QLabel(InstallDialog)
        self.shortcut_label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.shortcut_label.setObjectName("shortcut_label")
        self.install_dialog_layout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.shortcut_label)
        self.shortcut_check = QtWidgets.QCheckBox(InstallDialog)
        self.shortcut_check.setText("")
        self.shortcut_check.setObjectName("shortcut_check")
        self.install_dialog_layout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.shortcut_check)
        self.selectable_layout = QtWidgets.QVBoxLayout()
        self.selectable_layout.setObjectName("selectable_layout")
        self.install_dialog_layout.setLayout(4, QtWidgets.QFormLayout.SpanningRole, self.selectable_layout)
        self.advanced_layout = QtWidgets.QVBoxLayout()
        self.advanced_layout.setObjectName("advanced_layout")
        self.install_dialog_layout.setLayout(5, QtWidgets.QFormLayout.SpanningRole, self.advanced_layout)
        self.download_size_label = QtWidgets.QLabel(InstallDialog)
        self.download_size_label.setObjectName("download_size_label")
        self.install_dialog_layout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.download_size_label)
        self.download_size_text = QtWidgets.QLabel(InstallDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.download_size_text.setFont(font)
        self.download_size_text.setObjectName("download_size_text")
        self.install_dialog_layout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.download_size_text)
        self.install_size_label = QtWidgets.QLabel(InstallDialog)
        self.install_size_label.setObjectName("install_size_label")
        self.install_dialog_layout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.install_size_label)
        self.install_size_text = QtWidgets.QLabel(InstallDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.install_size_text.setFont(font)
        self.install_size_text.setWordWrap(True)
        self.install_size_text.setObjectName("install_size_text")
        self.install_dialog_layout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.install_size_text)
        self.warning_label = QtWidgets.QLabel(InstallDialog)
        self.warning_label.setObjectName("warning_label")
        self.install_dialog_layout.setWidget(9, QtWidgets.QFormLayout.LabelRole, self.warning_label)
        self.warning_text = QtWidgets.QLabel(InstallDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.warning_text.setFont(font)
        self.warning_text.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.warning_text.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.warning_text.setWordWrap(True)
        self.warning_text.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.warning_text.setObjectName("warning_text")
        self.install_dialog_layout.setWidget(9, QtWidgets.QFormLayout.FieldRole, self.warning_text)
        self.button_layout = QtWidgets.QHBoxLayout()
        self.button_layout.setObjectName("button_layout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.button_layout.addItem(spacerItem)
        self.cancel_button = QtWidgets.QPushButton(InstallDialog)
        self.cancel_button.setObjectName("cancel_button")
        self.button_layout.addWidget(self.cancel_button)
        self.verify_button = QtWidgets.QPushButton(InstallDialog)
        self.verify_button.setObjectName("verify_button")
        self.button_layout.addWidget(self.verify_button)
        self.install_button = QtWidgets.QPushButton(InstallDialog)
        self.install_button.setObjectName("install_button")
        self.button_layout.addWidget(self.install_button)
        self.install_dialog_layout.setLayout(10, QtWidgets.QFormLayout.SpanningRole, self.button_layout)
        self.avail_space_lbl = QtWidgets.QLabel(InstallDialog)
        self.avail_space_lbl.setObjectName("avail_space_lbl")
        self.install_dialog_layout.setWidget(8, QtWidgets.QFormLayout.LabelRole, self.avail_space_lbl)
        self.avail_space = QtWidgets.QLabel(InstallDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.avail_space.setFont(font)
        self.avail_space.setText("")
        self.avail_space.setObjectName("avail_space")
        self.install_dialog_layout.setWidget(8, QtWidgets.QFormLayout.FieldRole, self.avail_space)

        self.retranslateUi(InstallDialog)

    def retranslateUi(self, InstallDialog):
        _translate = QtCore.QCoreApplication.translate
        self.install_dialog_label.setText(_translate("InstallDialog", "error"))
        self.install_dir_label.setText(_translate("InstallDialog", "Install directory"))
        self.platform_label.setText(_translate("InstallDialog", "Platform"))
        self.shortcut_label.setText(_translate("InstallDialog", "Create shortcut"))
        self.download_size_label.setText(_translate("InstallDialog", "Download size"))
        self.download_size_text.setText(_translate("InstallDialog", "Click verify..."))
        self.install_size_label.setText(_translate("InstallDialog", "Total install size"))
        self.install_size_text.setText(_translate("InstallDialog", "Click verify..."))
        self.warning_label.setText(_translate("InstallDialog", "Warning"))
        self.warning_text.setText(_translate("InstallDialog", "None"))
        self.cancel_button.setText(_translate("InstallDialog", "Cancel"))
        self.verify_button.setText(_translate("InstallDialog", "Verify"))
        self.install_button.setText(_translate("InstallDialog", "Install"))
        self.avail_space_lbl.setText(_translate("InstallDialog", "Available space"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    InstallDialog = QtWidgets.QDialog()
    ui = Ui_InstallDialog()
    ui.setupUi(InstallDialog)
    InstallDialog.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/downloads/queue_base_widget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_QueueBaseWidget(object):
    def setupUi(self, QueueBaseWidget):
        QueueBaseWidget.setObjectName("QueueBaseWidget")
        QueueBaseWidget.resize(354, 72)
        QueueBaseWidget.setWindowTitle("QueueBaseWidget")
        QueueBaseWidget.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.queue_base_layout = QtWidgets.QHBoxLayout(QueueBaseWidget)
        self.queue_base_layout.setContentsMargins(0, 0, 0, 0)
        self.queue_base_layout.setObjectName("queue_base_layout")
        self.info_layout = QtWidgets.QHBoxLayout()
        self.info_layout.setObjectName("info_layout")
        self.queue_base_layout.addLayout(self.info_layout)
        self.move_buttons = QtWidgets.QWidget(QueueBaseWidget)
        self.move_buttons.setObjectName("move_buttons")
        self.move_buttons_layout = QtWidgets.QVBoxLayout(self.move_buttons)
        self.move_buttons_layout.setContentsMargins(0, 0, 0, 0)
        self.move_buttons_layout.setObjectName("move_buttons_layout")
        self.move_up_button = QtWidgets.QPushButton(self.move_buttons)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.move_up_button.sizePolicy().hasHeightForWidth())
        self.move_up_button.setSizePolicy(sizePolicy)
        self.move_up_button.setText("")
        self.move_up_button.setProperty("isSquare", 1)
        self.move_up_button.setObjectName("move_up_button")
        self.move_buttons_layout.addWidget(self.move_up_button, 0, QtCore.Qt.AlignLeft|QtCore.Qt.AlignTop)
        self.move_down_button = QtWidgets.QPushButton(self.move_buttons)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.move_down_button.sizePolicy().hasHeightForWidth())
        self.move_down_button.setSizePolicy(sizePolicy)
        self.move_down_button.setText("")
        self.move_down_button.setProperty("isSquare", 1)
        self.move_down_button.setObjectName("move_down_button")
        self.move_buttons_layout.addWidget(self.move_down_button, 0, QtCore.Qt.AlignLeft|QtCore.Qt.AlignBottom)
        self.queue_base_layout.addWidget(self.move_buttons)
        self.queue_buttons = QtWidgets.QWidget(QueueBaseWidget)
        self.queue_buttons.setObjectName("queue_buttons")
        self.queue_buttons_layout = QtWidgets.QVBoxLayout(self.queue_buttons)
        self.queue_buttons_layout.setContentsMargins(0, 0, 0, 0)
        self.queue_buttons_layout.setObjectName("queue_buttons_layout")
        self.force_button = QtWidgets.QPushButton(self.queue_buttons)
        self.force_button.setObjectName("force_button")
        self.queue_buttons_layout.addWidget(self.force_button)
        self.remove_button = QtWidgets.QPushButton(self.queue_buttons)
        self.remove_button.setObjectName("remove_button")
        self.queue_buttons_layout.addWidget(self.remove_button, 0, QtCore.Qt.AlignTop)
        self.queue_base_layout.addWidget(self.queue_buttons)
        self.update_buttons = QtWidgets.QWidget(QueueBaseWidget)
        self.update_buttons.setObjectName("update_buttons")
        self.update_buttons_layout = QtWidgets.QVBoxLayout(self.update_buttons)
        self.update_buttons_layout.setContentsMargins(0, 0, 0, 0)
        self.update_buttons_layout.setObjectName("update_buttons_layout")
        self.update_button = QtWidgets.QPushButton(self.update_buttons)
        self.update_button.setObjectName("update_button")
        self.update_buttons_layout.addWidget(self.update_button, 0, QtCore.Qt.AlignTop)
        self.settings_button = QtWidgets.QPushButton(self.update_buttons)
        self.settings_button.setObjectName("settings_button")
        self.update_buttons_layout.addWidget(self.settings_button, 0, QtCore.Qt.AlignTop)
        self.queue_base_layout.addWidget(self.update_buttons)
        self.queue_base_layout.setStretch(0, 1)

        self.retranslateUi(QueueBaseWidget)

    def retranslateUi(self, QueueBaseWidget):
        _translate = QtCore.QCoreApplication.translate
        self.force_button.setText(_translate("QueueBaseWidget", "Download now"))
        self.remove_button.setText(_translate("QueueBaseWidget", "Remove from queue"))
        self.update_button.setText(_translate("QueueBaseWidget", "Update game"))
        self.settings_button.setText(_translate("QueueBaseWidget", "Update with settings"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    QueueBaseWidget = QtWidgets.QFrame()
    ui = Ui_QueueBaseWidget()
    ui.setupUi(QueueBaseWidget)
    QueueBaseWidget.show()
    sys.exit(app.exec_())



from PyQt5.QtCore import Qt, QPropertyAnimation, QEasingCurve, QSize
from PyQt5.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QGraphicsOpacityEffect,
    QSpacerItem,
    QSizePolicy,
    QHBoxLayout,
    QLabel,
    QPushButton,
)

from rare.utils.misc import icon, widget_object_name
from rare.widgets.elide_label import ElideLabel


class IconWidget(object):
    def __init__(self):
        self._effect = None
        self._animation: QPropertyAnimation = None

        self.status_label: ElideLabel = None
        self.mini_widget: QWidget = None
        self.mini_effect: QGraphicsOpacityEffect = None
        self.title_label: QLabel = None
        self.tooltip_label: ElideLabel = None
        self.launch_btn: QPushButton = None
        self.install_btn: QPushButton = None

    def setupUi(self, widget: QWidget):
        # information at top
        self.status_label = ElideLabel(parent=widget)
        self.status_label.setObjectName(f"{type(self).__name__}StatusLabel")
        self.status_label.setFixedHeight(False)
        self.status_label.setContentsMargins(6, 6, 6, 6)
        self.status_label.setAutoFillBackground(False)

        # on-hover popup
        self.mini_widget = QWidget(parent=widget)
        self.mini_widget.setObjectName(f"{type(self).__name__}MiniWidget")
        self.mini_widget.setFixedHeight(widget.height() // 3)

        self.mini_effect = QGraphicsOpacityEffect(self.mini_widget)
        self.mini_widget.setGraphicsEffect(self.mini_effect)

        # game title
        self.title_label = QLabel(parent=self.mini_widget)
        self.title_label.setObjectName(f"{type(self).__name__}TitleLabel")
        self.title_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.title_label.setAlignment(Qt.AlignVCenter)
        self.title_label.setAutoFillBackground(False)
        self.title_label.setWordWrap(True)

        # information below title
        self.tooltip_label = ElideLabel(parent=self.mini_widget)
        self.tooltip_label.setObjectName(f"{type(self).__name__}TooltipLabel")
        self.tooltip_label.setAutoFillBackground(False)

        # play button
        self.launch_btn = QPushButton(parent=self.mini_widget)
        self.launch_btn.setObjectName(f"{type(self).__name__}Button")
        self.launch_btn.setIcon(icon("ei.play-alt", color="white"))
        self.launch_btn.setIconSize(QSize(20, 20))
        self.launch_btn.setFixedSize(QSize(widget.width() // 4, widget.width() // 4))

        self.install_btn = QPushButton(parent=self.mini_widget)
        self.install_btn.setObjectName(f"{type(self).__name__}Button")
        self.install_btn.setIcon(icon("ri.install-fill", color="white"))
        self.install_btn.setIconSize(QSize(20, 20))
        self.install_btn.setFixedSize(QSize(widget.width() // 4, widget.width() // 4))

        # lk: do not focus on button
        # When the button gets clicked on, it receives keyboard focus. Disabling the button
        # afterwards leads to `focusNextChild` getting called. This makes the scrollarea
        # trying to ensure that `nextChild` is visible, essentially scrolling to a random widget
        self.launch_btn.setFocusPolicy(Qt.NoFocus)
        self.install_btn.setFocusPolicy(Qt.NoFocus)

        # Create layouts
        # layout on top of the image, holds the status label, a spacer item and the mini widget
        image_layout = QVBoxLayout()
        image_layout.setContentsMargins(2, 2, 2, 2)

        # layout for the mini widget, holds the top row and the info label
        mini_layout = QVBoxLayout()
        mini_layout.setSpacing(0)

        # layout for the top row, holds the title and the launch button
        row_layout = QHBoxLayout()
        row_layout.setSpacing(0)
        row_layout.setAlignment(Qt.AlignTop)

        # Layout the widgets
        # (from inner to outer)
        row_layout.addWidget(self.title_label, stretch=2)
        row_layout.addWidget(self.launch_btn)
        row_layout.addWidget(self.install_btn)
        mini_layout.addLayout(row_layout, stretch=2)
        mini_layout.addWidget(self.tooltip_label)
        self.mini_widget.setLayout(mini_layout)

        image_layout.addWidget(self.status_label)
        image_layout.addSpacerItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Expanding))
        image_layout.addWidget(self.mini_widget)
        widget.setLayout(image_layout)

        widget.setContextMenuPolicy(Qt.ActionsContextMenu)
        widget.leaveEvent(None)

        self.translateUi(widget)

    def translateUi(self, widget: QWidget):
        pass

    def enterAnimation(self, widget: QWidget):
        self._animation = QPropertyAnimation(self.mini_effect, b"opacity")
        self._animation.setDuration(250)
        self._animation.setStartValue(0)
        self._animation.setEndValue(1)
        self._animation.setEasingCurve(QEasingCurve.InSine)
        self._animation.start(QPropertyAnimation.DeleteWhenStopped)

    def leaveAnimation(self, widget: QWidget):
        self._animation = QPropertyAnimation(self.mini_effect, b"opacity")
        self._animation.setDuration(150)
        self._animation.setStartValue(1)
        self._animation.setEndValue(0)
        self._animation.setEasingCurve(QEasingCurve.OutSine)
        self._animation.start(QPropertyAnimation.DeleteWhenStopped)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/games/game_info/game_info.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_GameInfo(object):
    def setupUi(self, GameInfo):
        GameInfo.setObjectName("GameInfo")
        GameInfo.resize(600, 404)
        self.main_layout = QtWidgets.QHBoxLayout(GameInfo)
        self.main_layout.setObjectName("main_layout")
        self.left_layout = QtWidgets.QVBoxLayout()
        self.left_layout.setObjectName("left_layout")
        self.tags_group = QtWidgets.QGroupBox(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tags_group.sizePolicy().hasHeightForWidth())
        self.tags_group.setSizePolicy(sizePolicy)
        self.tags_group.setObjectName("tags_group")
        self.tags_layout = QtWidgets.QGridLayout(self.tags_group)
        self.tags_layout.setHorizontalSpacing(0)
        self.tags_layout.setObjectName("tags_layout")
        self.completed_check = QtWidgets.QCheckBox(self.tags_group)
        self.completed_check.setObjectName("completed_check")
        self.tags_layout.addWidget(self.completed_check, 3, 0, 1, 2)
        self.hidden_check = QtWidgets.QCheckBox(self.tags_group)
        self.hidden_check.setObjectName("hidden_check")
        self.tags_layout.addWidget(self.hidden_check, 0, 0, 1, 2)
        self.custom1_edit = QtWidgets.QLineEdit(self.tags_group)
        self.custom1_edit.setObjectName("custom1_edit")
        self.tags_layout.addWidget(self.custom1_edit, 4, 1, 1, 1)
        self.favorites_check = QtWidgets.QCheckBox(self.tags_group)
        self.favorites_check.setObjectName("favorites_check")
        self.tags_layout.addWidget(self.favorites_check, 1, 0, 1, 2)
        self.custom1_check = QtWidgets.QCheckBox(self.tags_group)
        self.custom1_check.setText("")
        self.custom1_check.setObjectName("custom1_check")
        self.tags_layout.addWidget(self.custom1_check, 4, 0, 1, 1)
        self.backlog_check = QtWidgets.QCheckBox(self.tags_group)
        self.backlog_check.setObjectName("backlog_check")
        self.tags_layout.addWidget(self.backlog_check, 2, 0, 1, 2)
        self.custom2_check = QtWidgets.QCheckBox(self.tags_group)
        self.custom2_check.setText("")
        self.custom2_check.setObjectName("custom2_check")
        self.tags_layout.addWidget(self.custom2_check, 5, 0, 1, 1)
        self.custom2_edit = QtWidgets.QLineEdit(self.tags_group)
        self.custom2_edit.setObjectName("custom2_edit")
        self.tags_layout.addWidget(self.custom2_edit, 5, 1, 1, 1)
        self.left_layout.addWidget(self.tags_group)
        self.main_layout.addLayout(self.left_layout)
        self.right_layout = QtWidgets.QVBoxLayout()
        self.right_layout.setObjectName("right_layout")
        self.info_layout = QtWidgets.QFormLayout()
        self.info_layout.setLabelAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.info_layout.setContentsMargins(6, 6, 6, 6)
        self.info_layout.setSpacing(12)
        self.info_layout.setObjectName("info_layout")
        self.lbl_dev = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_dev.sizePolicy().hasHeightForWidth())
        self.lbl_dev.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_dev.setFont(font)
        self.lbl_dev.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_dev.setObjectName("lbl_dev")
        self.info_layout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.lbl_dev)
        self.dev = QtWidgets.QLabel(GameInfo)
        self.dev.setText("error")
        self.dev.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.dev.setObjectName("dev")
        self.info_layout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.dev)
        self.lbl_app_name = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_app_name.sizePolicy().hasHeightForWidth())
        self.lbl_app_name.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_app_name.setFont(font)
        self.lbl_app_name.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_app_name.setObjectName("lbl_app_name")
        self.info_layout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.lbl_app_name)
        self.app_name = QtWidgets.QLabel(GameInfo)
        self.app_name.setText("error")
        self.app_name.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.app_name.setObjectName("app_name")
        self.info_layout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.app_name)
        self.lbl_version = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_version.sizePolicy().hasHeightForWidth())
        self.lbl_version.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_version.setFont(font)
        self.lbl_version.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_version.setObjectName("lbl_version")
        self.info_layout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.lbl_version)
        self.version = QtWidgets.QLabel(GameInfo)
        self.version.setText("error")
        self.version.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.version.setObjectName("version")
        self.info_layout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.version)
        self.lbl_grade = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_grade.sizePolicy().hasHeightForWidth())
        self.lbl_grade.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_grade.setFont(font)
        self.lbl_grade.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_grade.setObjectName("lbl_grade")
        self.info_layout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.lbl_grade)
        self.grade = QtWidgets.QLabel(GameInfo)
        self.grade.setText("error")
        self.grade.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.grade.setObjectName("grade")
        self.info_layout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.grade)
        self.lbl_install_size = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_install_size.sizePolicy().hasHeightForWidth())
        self.lbl_install_size.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_install_size.setFont(font)
        self.lbl_install_size.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_install_size.setObjectName("lbl_install_size")
        self.info_layout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.lbl_install_size)
        self.install_size = QtWidgets.QLabel(GameInfo)
        self.install_size.setText("error")
        self.install_size.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.install_size.setObjectName("install_size")
        self.info_layout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.install_size)
        self.lbl_install_path = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_install_path.sizePolicy().hasHeightForWidth())
        self.lbl_install_path.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_install_path.setFont(font)
        self.lbl_install_path.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_install_path.setObjectName("lbl_install_path")
        self.info_layout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.lbl_install_path)
        self.install_path = QtWidgets.QLabel(GameInfo)
        self.install_path.setText("error")
        self.install_path.setWordWrap(True)
        self.install_path.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByMouse)
        self.install_path.setObjectName("install_path")
        self.info_layout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.install_path)
        self.lbl_platform = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_platform.sizePolicy().hasHeightForWidth())
        self.lbl_platform.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_platform.setFont(font)
        self.lbl_platform.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_platform.setObjectName("lbl_platform")
        self.info_layout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.lbl_platform)
        self.platform = QtWidgets.QLabel(GameInfo)
        self.platform.setText("error")
        self.platform.setObjectName("platform")
        self.info_layout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.platform)
        self.lbl_game_actions = QtWidgets.QLabel(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lbl_game_actions.sizePolicy().hasHeightForWidth())
        self.lbl_game_actions.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.lbl_game_actions.setFont(font)
        self.lbl_game_actions.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.lbl_game_actions.setObjectName("lbl_game_actions")
        self.info_layout.setWidget(7, QtWidgets.QFormLayout.LabelRole, self.lbl_game_actions)
        self.game_actions_stack = QtWidgets.QStackedWidget(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.game_actions_stack.sizePolicy().hasHeightForWidth())
        self.game_actions_stack.setSizePolicy(sizePolicy)
        self.game_actions_stack.setMinimumSize(QtCore.QSize(250, 0))
        self.game_actions_stack.setMaximumSize(QtCore.QSize(250, 16777215))
        self.game_actions_stack.setObjectName("game_actions_stack")
        self.installed_page = QtWidgets.QWidget()
        self.installed_page.setObjectName("installed_page")
        self.installed_layout = QtWidgets.QVBoxLayout(self.installed_page)
        self.installed_layout.setContentsMargins(0, 0, 0, 0)
        self.installed_layout.setObjectName("installed_layout")
        self.modify_button = QtWidgets.QPushButton(self.installed_page)
        self.modify_button.setObjectName("modify_button")
        self.installed_layout.addWidget(self.modify_button)
        self.verify_stack = QtWidgets.QStackedWidget(self.installed_page)
        self.verify_stack.setObjectName("verify_stack")
        self.verify_button_page = QtWidgets.QWidget()
        self.verify_button_page.setObjectName("verify_button_page")
        self.verify_page_layout = QtWidgets.QHBoxLayout(self.verify_button_page)
        self.verify_page_layout.setContentsMargins(0, 0, 0, 0)
        self.verify_page_layout.setSpacing(0)
        self.verify_page_layout.setObjectName("verify_page_layout")
        self.verify_button = QtWidgets.QPushButton(self.verify_button_page)
        self.verify_button.setObjectName("verify_button")
        self.verify_page_layout.addWidget(self.verify_button)
        self.verify_stack.addWidget(self.verify_button_page)
        self.verify_progress_page = QtWidgets.QWidget()
        self.verify_progress_page.setObjectName("verify_progress_page")
        self.verify_progress_layout = QtWidgets.QHBoxLayout(self.verify_progress_page)
        self.verify_progress_layout.setContentsMargins(0, 0, 0, 0)
        self.verify_progress_layout.setSpacing(0)
        self.verify_progress_layout.setObjectName("verify_progress_layout")
        self.verify_progress = QtWidgets.QProgressBar(self.verify_progress_page)
        self.verify_progress.setProperty("value", 24)
        self.verify_progress.setObjectName("verify_progress")
        self.verify_progress_layout.addWidget(self.verify_progress)
        self.verify_stack.addWidget(self.verify_progress_page)
        self.installed_layout.addWidget(self.verify_stack)
        self.repair_button = QtWidgets.QPushButton(self.installed_page)
        self.repair_button.setObjectName("repair_button")
        self.installed_layout.addWidget(self.repair_button)
        self.move_stack = QtWidgets.QStackedWidget(self.installed_page)
        self.move_stack.setObjectName("move_stack")
        self.move_button_page = QtWidgets.QWidget()
        self.move_button_page.setObjectName("move_button_page")
        self.move_button_layout = QtWidgets.QHBoxLayout(self.move_button_page)
        self.move_button_layout.setContentsMargins(0, 0, 0, 0)
        self.move_button_layout.setSpacing(0)
        self.move_button_layout.setObjectName("move_button_layout")
        self.move_button = QtWidgets.QToolButton(self.move_button_page)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.move_button.sizePolicy().hasHeightForWidth())
        self.move_button.setSizePolicy(sizePolicy)
        self.move_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.move_button.setToolButtonStyle(QtCore.Qt.ToolButtonTextOnly)
        self.move_button.setArrowType(QtCore.Qt.DownArrow)
        self.move_button.setObjectName("move_button")
        self.move_button_layout.addWidget(self.move_button)
        self.move_stack.addWidget(self.move_button_page)
        self.move_progress_page = QtWidgets.QWidget()
        self.move_progress_page.setObjectName("move_progress_page")
        self.move_progress_layout = QtWidgets.QHBoxLayout(self.move_progress_page)
        self.move_progress_layout.setContentsMargins(0, 0, 0, 0)
        self.move_progress_layout.setSpacing(0)
        self.move_progress_layout.setObjectName("move_progress_layout")
        self.move_progress = QtWidgets.QProgressBar(self.move_progress_page)
        self.move_progress.setProperty("value", 24)
        self.move_progress.setObjectName("move_progress")
        self.move_progress_layout.addWidget(self.move_progress)
        self.move_stack.addWidget(self.move_progress_page)
        self.installed_layout.addWidget(self.move_stack)
        self.uninstall_button = QtWidgets.QPushButton(self.installed_page)
        self.uninstall_button.setObjectName("uninstall_button")
        self.installed_layout.addWidget(self.uninstall_button)
        self.game_actions_stack.addWidget(self.installed_page)
        self.uninstalled_page = QtWidgets.QWidget()
        self.uninstalled_page.setObjectName("uninstalled_page")
        self.uninstalled_layout = QtWidgets.QVBoxLayout(self.uninstalled_page)
        self.uninstalled_layout.setContentsMargins(0, 0, 0, 0)
        self.uninstalled_layout.setObjectName("uninstalled_layout")
        self.install_button = QtWidgets.QPushButton(self.uninstalled_page)
        self.install_button.setObjectName("install_button")
        self.uninstalled_layout.addWidget(self.install_button)
        self.import_button = QtWidgets.QPushButton(self.uninstalled_page)
        self.import_button.setObjectName("import_button")
        self.uninstalled_layout.addWidget(self.import_button)
        spacerItem = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.uninstalled_layout.addItem(spacerItem)
        self.game_actions_stack.addWidget(self.uninstalled_page)
        self.info_layout.setWidget(7, QtWidgets.QFormLayout.FieldRole, self.game_actions_stack)
        self.right_layout.addLayout(self.info_layout)
        self.requirements_group = QtWidgets.QFrame(GameInfo)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.requirements_group.sizePolicy().hasHeightForWidth())
        self.requirements_group.setSizePolicy(sizePolicy)
        self.requirements_group.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.requirements_group.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.requirements_group.setObjectName("requirements_group")
        self.requirements_layout = QtWidgets.QHBoxLayout(self.requirements_group)
        self.requirements_layout.setContentsMargins(0, 0, 0, 0)
        self.requirements_layout.setObjectName("requirements_layout")
        self.right_layout.addWidget(self.requirements_group)
        self.main_layout.addLayout(self.right_layout)
        self.main_layout.setStretch(1, 1)

        self.retranslateUi(GameInfo)
        self.game_actions_stack.setCurrentIndex(0)
        self.verify_stack.setCurrentIndex(0)
        self.move_stack.setCurrentIndex(0)

    def retranslateUi(self, GameInfo):
        _translate = QtCore.QCoreApplication.translate
        GameInfo.setWindowTitle(_translate("GameInfo", "Game Info"))
        self.tags_group.setTitle(_translate("GameInfo", "Tags"))
        self.completed_check.setText(_translate("GameInfo", "Completed"))
        self.hidden_check.setText(_translate("GameInfo", "Hidden"))
        self.favorites_check.setText(_translate("GameInfo", "Favorites"))
        self.backlog_check.setText(_translate("GameInfo", "Backlog"))
        self.lbl_dev.setText(_translate("GameInfo", "Developer"))
        self.lbl_app_name.setText(_translate("GameInfo", "Application Name"))
        self.lbl_version.setText(_translate("GameInfo", "Version"))
        self.lbl_grade.setText(_translate("GameInfo", "ProtonDB Grade"))
        self.lbl_install_size.setText(_translate("GameInfo", "Installation Size"))
        self.lbl_install_path.setText(_translate("GameInfo", "Installation Path"))
        self.lbl_platform.setText(_translate("GameInfo", "Platform"))
        self.lbl_game_actions.setText(_translate("GameInfo", "Actions"))
        self.modify_button.setText(_translate("GameInfo", "Modify Installation"))
        self.verify_button.setText(_translate("GameInfo", "Verify Installation"))
        self.repair_button.setText(_translate("GameInfo", "Repair Installation"))
        self.move_button.setText(_translate("GameInfo", "Move Installation"))
        self.uninstall_button.setText(_translate("GameInfo", "Uninstall Game"))
        self.install_button.setText(_translate("GameInfo", "Install Game"))
        self.import_button.setText(_translate("GameInfo", "Import Game"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    GameInfo = QtWidgets.QWidget()
    ui = Ui_GameInfo()
    ui.setupUi(GameInfo)
    GameInfo.show()
    sys.exit(app.exec_())



import datetime
import platform
from logging import getLogger
from typing import Union, Optional

from PyQt5.QtCore import pyqtSignal, QSettings, pyqtSlot, QThreadPool, Qt
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import (
    QWidget,
    QMessageBox, QScrollArea, QVBoxLayout, QSizePolicy,
)

from rare.components.dialogs.install_dialog import InstallDialog
from rare.components.dialogs.uninstall_dialog import UninstallDialog
from rare.lgndr.models.downloading import UIUpdate
from rare.models.game import RareGame
from rare.models.install import InstallOptionsModel, InstallQueueItemModel, UninstallOptionsModel
from rare.shared import RareCore
from rare.shared.workers.install_info import InstallInfoWorker
from rare.shared.workers.uninstall import UninstallWorker
from rare.utils.misc import format_size
from rare.utils.paths import create_desktop_link, desktop_links_supported
from .download import DownloadWidget
from .groups import UpdateGroup, QueueGroup
from .thread import DlThread, DlResultModel, DlResultCode
from .widgets import UpdateWidget, QueueWidget

logger = getLogger("Download")


def get_time(seconds: Union[int, float]) -> str:
    return str(datetime.timedelta(seconds=seconds))


class DownloadsTab(QWidget):
    # int: number of updates
    update_title = pyqtSignal(int)

    def __init__(self, parent=None):
        super(DownloadsTab, self).__init__(parent=parent)
        self.rcore = RareCore.instance()
        self.core = RareCore.instance().core()
        self.signals = RareCore.instance().signals()
        self.args = RareCore.instance().args()

        self.__thread: Optional[DlThread] = None

        layout = QVBoxLayout(self)

        self.download_widget = DownloadWidget(self)
        self.download_widget.ui.kill_button.clicked.connect(self.stop_download)
        layout.addWidget(self.download_widget)

        self.queue_scrollarea = QScrollArea(self)
        self.queue_scrollarea.setWidgetResizable(True)
        self.queue_scrollarea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.queue_scrollarea.setFrameStyle(QScrollArea.Plain | QScrollArea.NoFrame)
        layout.addWidget(self.queue_scrollarea)

        queue_contents = QWidget(self.queue_scrollarea)
        queue_contents.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        self.queue_scrollarea.setWidget(queue_contents)

        queue_contents_layout = QVBoxLayout(queue_contents)
        queue_contents_layout.setContentsMargins(0, 0, 3, 0)

        self.queue_group = QueueGroup(self)
        self.queue_group.update_count.connect(self.update_queues_count)
        self.queue_group.removed.connect(self.__on_queue_removed)
        self.queue_group.force.connect(self.__on_queue_force)
        queue_contents_layout.addWidget(self.queue_group)

        self.updates_group = UpdateGroup(self)
        self.updates_group.update_count.connect(self.update_queues_count)
        self.updates_group.enqueue.connect(self.__get_install_options)
        queue_contents_layout.addWidget(self.updates_group)

        self.__check_updates()
        self.__reset_download()

        self.signals.game.install.connect(self.__get_install_options)
        self.signals.game.uninstall.connect(self.__get_uninstall_options)
        self.signals.download.enqueue.connect(self.__add_update)
        self.signals.download.dequeue.connect(self.__remove_update)

        self.__forced_item: Optional[InstallQueueItemModel] = None
        self.__omit_requeue = False

    @pyqtSlot()
    @pyqtSlot(int)
    def update_queues_count(self):
        count = self.updates_group.count() + self.queue_group.count() + (1 if self.is_download_active else 0)
        self.update_title.emit(count)

    @property
    def is_download_active(self):
        return self.__thread is not None

    def __check_updates(self):
        for rgame in self.rcore.updates:
            self.__add_update(rgame)

    @pyqtSlot(str)
    @pyqtSlot(RareGame)
    def __add_update(self, update: Union[str, RareGame]):
        if isinstance(update, str):
            update = self.rcore.get_game(update)
        if update.metadata.auto_update or QSettings().value("auto_update", False, bool):
            self.__get_install_options(
                InstallOptionsModel(app_name=update.app_name, update=True, silent=True)
            )
        else:
            self.updates_group.append(update.game, update.igame)
        self.update_queues_count()

    @pyqtSlot(str)
    def __remove_update(self, app_name):
        if self.__thread and self.__thread.item.options.app_name == app_name:
            self.stop_download(omit_queue=True)
        if self.queue_group.contains(app_name):
            self.queue_group.remove(app_name)
        if self.updates_group.contains(app_name):
            self.updates_group.remove(app_name)

    @pyqtSlot(str)
    def __on_queue_removed(self, app_name: str):
        """
        Handle removing a queued item.
        If the item exists in the updates (it means a repair was removed), re-enable the buttons.
        If it doesn't exist in the updates, recreate the widget.
        :param app_name:
        :return:
        """
        rgame = self.rcore.get_game(app_name)
        rgame.state = RareGame.State.IDLE
        if self.updates_group.contains(app_name):
            self.updates_group.set_widget_enabled(app_name, True)
        else:
            if rgame.is_installed and rgame.has_update:
                self.__add_update(app_name)

    @pyqtSlot(InstallQueueItemModel)
    def __on_queue_force(self, item: InstallQueueItemModel):
        if self.__thread:
            self.stop_download()
            self.__forced_item = item
        else:
            self.__start_download(item)

    def stop_download(self, omit_queue=False):
        """
        Stops the active download, by optionally skipping the queue

        :param omit_queue: bool
            If `True`, the stopped download won't be added back to the queue.
            Defaults to `False`
        :return:
        """
        self.__thread.kill()
        self.download_widget.ui.kill_button.setEnabled(False)
        # lk: if we are exiting Rare, wait for thread to finish
        # `self.on_exit` control whether we try to add the download
        # back in the queue. If we are on exit we wait for the thread
        # to finish, we do not care about handling the result really
        self.__omit_requeue = omit_queue
        if omit_queue:
            self.__thread.wait()

    def __refresh_download(self, item: InstallQueueItemModel):
        worker = InstallInfoWorker(self.core, item.options)
        worker.signals.result.connect(
            lambda d: self.__start_download(InstallQueueItemModel(options=item.options, download=d))
        )
        worker.signals.failed.connect(
            lambda m: logger.error(f"Failed to refresh download for {item.options.app_name} with error: {m}")
        )
        worker.signals.finished.connect(
            lambda: logger.info(f"Download refresh worker finished for {item.options.app_name}")
        )
        QThreadPool.globalInstance().start(worker)

    def __start_download(self, item: InstallQueueItemModel):
        rgame = self.rcore.get_game(item.options.app_name)
        if not rgame.state == RareGame.State.DOWNLOADING:
            logger.error(
                f"Can't start download {item.options.app_name}"
                f"due to incompatible state {RareGame.State(rgame.state).name}"
            )
            # lk: invalidate the queue item in case the game was uninstalled
            self.__requeue_download(InstallQueueItemModel(options=item.options))
            return
        if item.expired:
            self.__refresh_download(item)
            return
        dl_thread = DlThread(item, self.rcore.get_game(item.options.app_name), self.core, self.args.debug)
        dl_thread.result.connect(self.__on_download_result)
        dl_thread.progress.connect(self.__on_download_progress)
        dl_thread.finished.connect(dl_thread.deleteLater)
        dl_thread.start()
        self.__thread = dl_thread
        self.download_widget.ui.kill_button.setDisabled(False)
        self.download_widget.ui.dl_name.setText(item.download.game.app_title)
        self.download_widget.setPixmap(
            RareCore.instance().image_manager().get_pixmap(rgame.app_name, True)
        )

    @pyqtSlot(UIUpdate, object)
    def __on_download_progress(self, ui_update: UIUpdate, dl_size: int):
        self.download_widget.ui.progress_bar.setValue(int(ui_update.progress))
        self.download_widget.ui.dl_speed.setText(f"{format_size(ui_update.download_compressed_speed)}/s")
        self.download_widget.ui.cache_used.setText(
            f"{format_size(ui_update.cache_usage) if ui_update.cache_usage > 1023 else '0KB'}"
        )
        self.download_widget.ui.downloaded.setText(
            f"{format_size(ui_update.total_downloaded)} / {format_size(dl_size)}"
        )
        self.download_widget.ui.time_left.setText(get_time(ui_update.estimated_time_left))

    def __requeue_download(self, item: InstallQueueItemModel):
        rgame = self.rcore.get_game(item.options.app_name)
        rgame.state = RareGame.State.DOWNLOADING
        self.queue_group.push_front(item, rgame.igame)
        logger.info(f"Re-queued download for {rgame.app_name} ({rgame.app_title})")

    @pyqtSlot(DlResultModel)
    def __on_download_result(self, result: DlResultModel):
        if result.code == DlResultCode.FINISHED:
            logger.info(f"Download finished: {result.options.app_name}")
            if result.shortcut and desktop_links_supported():
                if not create_desktop_link(
                    app_name=result.options.app_name,
                    app_title=result.shortcut_title,
                    link_name=result.shortcut_name,
                    link_type="desktop",
                ):
                    # maybe add it to download summary, to show in finished downloads
                    logger.error(f"Failed to create desktop link on {platform.system()}")
                else:
                    logger.info(f"Created desktop link {result.shortcut_name} for {result.options.app_name}")

            if result.options.overlay:
                self.signals.application.overlay_installed.emit()
            else:
                self.signals.application.notify.emit(result.options.app_name)

            if self.updates_group.contains(result.options.app_name):
                self.updates_group.set_widget_enabled(result.options.app_name, True)

        elif result.code == DlResultCode.ERROR:
            logger.error(f"Download error: {result.options.app_name} ({result.message})")
            QMessageBox.warning(self, self.tr("Error"), self.tr("Download error: {}").format(result.message))

        elif result.code == DlResultCode.STOPPED:
            logger.info(f"Download stopped: {result.options.app_name}")
            if not self.__omit_requeue:
                self.__requeue_download(InstallQueueItemModel(options=result.options))
            else:
                return

        # lk: if we finished a repair, and we have a disabled update, re-enable it
        if self.updates_group.contains(result.options.app_name):
            self.updates_group.set_widget_enabled(result.options.app_name, True)

        if result.code == DlResultCode.FINISHED and self.queue_group.count():
            self.__start_download(self.queue_group.pop_front())
        elif result.code == DlResultCode.STOPPED and self.__forced_item:
            self.__start_download(self.__forced_item)
            self.__forced_item = None
        else:
            self.__reset_download()

    def __reset_download(self):
        self.download_widget.setPixmap(QPixmap())
        self.download_widget.ui.kill_button.setDisabled(True)
        self.download_widget.ui.dl_name.setText(self.tr("No active download"))
        self.download_widget.ui.progress_bar.setValue(0)
        self.download_widget.ui.dl_speed.setText("...")
        self.download_widget.ui.time_left.setText("...")
        self.download_widget.ui.cache_used.setText("...")
        self.download_widget.ui.downloaded.setText("...")
        self.__thread = None

    @pyqtSlot(InstallOptionsModel)
    def __get_install_options(self, options: InstallOptionsModel):
        rgame = self.rcore.get_game(options.app_name)
        rgame.state = RareGame.State.DOWNLOADING
        install_dialog = InstallDialog(
            rgame,
            options=options,
            parent=self,
        )
        install_dialog.result_ready.connect(self.__on_install_dialog_closed)
        install_dialog.execute()

    @pyqtSlot(InstallQueueItemModel)
    def __on_install_dialog_closed(self, item: InstallQueueItemModel):
        rgame = self.rcore.get_game(item.options.app_name)
        if item and not item.download.game.is_dlc and not item.download.analysis.dl_size:
            rgame.set_installed(True)
            rgame.state = RareGame.State.IDLE
            return
        if item:
            # lk: start update only if there is no other active thread and there is no queue
            if self.__thread is None and not self.queue_group.count():
                self.__start_download(item)
            else:
                rgame = self.rcore.get_game(item.options.app_name)
                self.queue_group.push_back(item, rgame.igame)
            # lk: Handle repairing into the current version
            # When we add something to the queue from repair, we might select to update or not
            # if we do select to update with repair, we can remove the widget from the updates groups
            # otherwise we disable it and keep it in the updates
            if self.updates_group.contains(item.options.app_name):
                if item.download.igame.version == self.updates_group.get_widget_version(item.options.app_name):
                    self.updates_group.remove(item.options.app_name)
                else:
                    self.updates_group.set_widget_enabled(item.options.app_name, False)
        else:
            if self.updates_group.contains(item.options.app_name):
                self.updates_group.set_widget_enabled(item.options.app_name, True)
            rgame.state = RareGame.State.IDLE

    @pyqtSlot(UninstallOptionsModel)
    def __get_uninstall_options(self, options: UninstallOptionsModel):
        rgame = self.rcore.get_game(options.app_name)
        rgame.state = RareGame.State.UNINSTALLING
        uninstall_dialog = UninstallDialog(
            rgame,
            options=options,
            parent=self,
        )
        uninstall_dialog.result_ready.connect(self.__on_uninstall_dialog_closed)
        uninstall_dialog.exec()

    @pyqtSlot(UninstallOptionsModel)
    def __on_uninstall_dialog_closed(self, options: UninstallOptionsModel):
        rgame = self.rcore.get_game(options.app_name)
        if options and options.accepted:
            rgame.set_installed(False)
            worker = UninstallWorker(self.core, rgame, options)
            worker.signals.result.connect(self.__on_uninstall_worker_result)
            QThreadPool.globalInstance().start(worker)
        else:
            rgame.state = RareGame.State.IDLE

    @pyqtSlot(RareGame, bool, str)
    def __on_uninstall_worker_result(self, rgame: RareGame, success: bool, message: str):
        if not success:
            QMessageBox.warning(None, self.tr("Uninstall - {}").format(rgame.app_title), message, QMessageBox.Close)
        rgame.state = RareGame.State.IDLE



from logging import getLogger
from typing import List

from PyQt5.QtCore import pyqtSignal, pyqtSlot, QSettings
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QSystemTrayIcon, QMenu, QAction, QApplication

from rare.shared import RareCore

logger = getLogger("TrayIcon")


class TrayIcon(QSystemTrayIcon):
    # none:
    show_app: pyqtSignal = pyqtSignal()
    # int: exit code
    exit_app: pyqtSignal = pyqtSignal(int)

    def __init__(self, parent=None):
        super(TrayIcon, self).__init__(parent=parent)
        self.__parent = parent
        self.rcore = RareCore.instance()
        self.core = RareCore.instance().core()
        self.settings = QSettings()

        self.setIcon(QIcon(":/images/Rare.png"))
        self.setVisible(True)
        self.setToolTip(QApplication.applicationName())

        self.menu = QMenu()

        self.show_action = QAction(QApplication.applicationName())
        self.show_action.triggered.connect(self.show_app)
        self.menu.addAction(self.show_action)

        self.menu.addSeparator()
        self.text_action = QAction("Quick launch")
        self.text_action.setEnabled(False)
        self.menu.addAction(self.text_action)

        # We need to reference this separator to add game actions before it
        self.separator = self.menu.addSeparator()
        self.exit_action = QAction(self.tr("Quit"))
        self.exit_action.triggered.connect(lambda: self.exit_app.emit(0))
        self.menu.addAction(self.exit_action)

        self.game_actions: List[QAction] = []
        self.update_actions()

        self.setContextMenu(self.menu)

        self.signals = RareCore.instance().signals()
        self.signals.game.uninstalled.connect(self.remove_button)
        self.signals.application.notify.connect(self.notify)
        self.signals.application.update_tray.connect(self.update_actions)

    def last_played(self) -> List:
        last_played = [game for game in self.rcore.games if (game.metadata and game.is_installed)]
        last_played.sort(key=lambda g: g.metadata.last_played, reverse=True)
        return last_played[0:5]

    @pyqtSlot(str)
    def notify(self, app_name: str):
        if self.settings.value("notification", True, bool):
            self.showMessage(
                self.tr("Download finished"),
                self.tr("Download finished. {} is playable now").format(
                    self.rcore.get_game(app_name).app_title
                ),
                self.Information,
                4000,
            )

    @pyqtSlot()
    def update_actions(self):
        for action in self.game_actions:
            action.deleteLater()
        self.game_actions.clear()
        for rgame in self.last_played():
            a = QAction(rgame.app_title)
            a.setProperty("app_name", rgame.app_name)
            a.triggered.connect(
                lambda: self.rcore.get_game(self.sender().property("app_name")).launch()
            )
            self.menu.insertAction(self.separator, a)
            self.game_actions.append(a)

    @pyqtSlot(str)
    def remove_button(self, app_name: str):
        if action := next((i for i in self.game_actions if i.property("app_name") == app_name), None):
            self.game_actions.remove(action)
            action.deleteLater()



from typing import Optional, List

from PyQt5.QtCore import Qt, pyqtSlot, pyqtSignal
from PyQt5.QtWidgets import QFrame, QMessageBox, QToolBox

from rare.models.game import RareGame
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton
from rare.ui.components.tabs.games.game_info.game_dlc import Ui_GameDlc
from rare.ui.components.tabs.games.game_info.game_dlc_widget import Ui_GameDlcWidget
from rare.widgets.image_widget import ImageWidget, ImageSize
from rare.widgets.side_tab import SideTabContents
from rare.utils.misc import widget_object_name, icon


class GameDlcWidget(QFrame):
    def __init__(self, rgame: RareGame, rdlc: RareGame, parent=None):
        super(GameDlcWidget, self).__init__(parent=parent)
        self.ui = Ui_GameDlcWidget()
        self.ui.setupUi(self)
        self.setObjectName(widget_object_name(self, rdlc.app_name))
        self.rgame = rgame
        self.rdlc = rdlc

        self.image = ImageWidget(self)
        self.image.setFixedSize(ImageSize.Icon)
        self.ui.dlc_layout.insertWidget(0, self.image)

        self.ui.dlc_name.setText(rdlc.app_title)
        self.ui.version.setText(rdlc.version)
        self.ui.app_name.setText(rdlc.app_name)

        self.image.setPixmap(rdlc.pixmap)

        self.__update()
        rdlc.signals.widget.update.connect(self.__update)

    @pyqtSlot()
    def __update(self):
        self.ui.action_button.setEnabled(self.rdlc.is_idle)


class InstalledGameDlcWidget(GameDlcWidget):
    uninstalled = pyqtSignal(RareGame)

    def __init__(self, rgame: RareGame, rdlc: RareGame, parent=None):
        super(InstalledGameDlcWidget, self).__init__(rgame=rgame, rdlc=rdlc, parent=parent)
        # lk: set object names for CSS properties
        self.ui.action_button.setObjectName("UninstallButton")
        self.ui.action_button.clicked.connect(self.uninstall_dlc)
        self.ui.action_button.setText(self.tr("Uninstall DLC"))
        self.ui.action_button.setIcon(icon("ri.uninstall-line"))
        # lk: don't reference `self.rdlc` here because the object has been deleted
        rdlc.signals.game.uninstalled.connect(self.__uninstalled)

    @pyqtSlot()
    def __uninstalled(self):
        self.uninstalled.emit(self.rdlc)

    def uninstall_dlc(self):
        self.rdlc.uninstall()


class AvailableGameDlcWidget(GameDlcWidget):
    installed = pyqtSignal(RareGame)

    def __init__(self, rgame: RareGame, rdlc: RareGame, parent=None):
        super(AvailableGameDlcWidget, self).__init__(rgame=rgame, rdlc=rdlc, parent=parent)
        # lk: set object names for CSS properties
        self.ui.action_button.setObjectName("InstallButton")
        self.ui.action_button.clicked.connect(self.install_dlc)
        self.ui.action_button.setText(self.tr("Install DLC"))
        self.ui.action_button.setIcon(icon("ri.install-line"))

        # lk: don't reference `self.rdlc` here because the object has been deleted
        rdlc.signals.game.installed.connect(self.__installed)

    @pyqtSlot()
    def __installed(self):
        self.installed.emit(self.rdlc)

    def install_dlc(self):
        if not self.rgame.is_installed:
            QMessageBox.warning(
                self,
                self.tr("Error"),
                self.tr("Base Game is not installed. Please install {} first").format(self.rgame.app_title),
            )
            return
        self.rdlc.install()


class GameDlc(QToolBox, SideTabContents):

    def __init__(self, parent=None):
        super(GameDlc, self).__init__(parent=parent)
        self.implements_scrollarea = True
        self.ui = Ui_GameDlc()
        self.ui.setupUi(self)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()

        self.rgame: Optional[RareGame] = None

    def list_installed(self) -> List[InstalledGameDlcWidget]:
        return self.ui.installed_dlc_container.findChildren(InstalledGameDlcWidget, options=Qt.FindDirectChildrenOnly)

    def list_available(self) -> List[AvailableGameDlcWidget]:
        return self.ui.available_dlc_container.findChildren(AvailableGameDlcWidget, options=Qt.FindDirectChildrenOnly)

    def get_installed(self, app_name: str) -> Optional[InstalledGameDlcWidget]:
        return self.ui.installed_dlc_container.findChild(
            InstalledGameDlcWidget,
            name=widget_object_name(InstalledGameDlcWidget, app_name),
            options=Qt.FindDirectChildrenOnly
        )

    def get_available(self, app_name: str) -> Optional[AvailableGameDlcWidget]:
        return self.ui.available_dlc_container.findChild(
            AvailableGameDlcWidget,
            name=widget_object_name(AvailableGameDlcWidget, app_name),
            options=Qt.FindDirectChildrenOnly
        )

    def update_installed_page(self):
        have_installed = bool(self.list_installed())
        self.ui.installed_dlc_label.setVisible(not have_installed)
        self.ui.installed_dlc_container.setVisible(have_installed)
        if not have_installed:
            self.setCurrentWidget(self.ui.available_dlc_page)

    def update_available_page(self):
        have_available = bool(self.list_available())
        self.ui.available_dlc_label.setVisible(not have_available)
        self.ui.available_dlc_container.setVisible(have_available)
        if not have_available:
            self.setCurrentWidget(self.ui.installed_dlc_page)

    def append_installed(self, rdlc: RareGame):
        self.ui.installed_dlc_label.setVisible(False)
        self.ui.installed_dlc_container.setVisible(True)
        a_widget: AvailableGameDlcWidget = self.get_available(rdlc.app_name)
        if a_widget is not None:
            self.ui.available_dlc_container.layout().removeWidget(a_widget)
            a_widget.deleteLater()
        i_widget: InstalledGameDlcWidget = InstalledGameDlcWidget(
            self.rgame, rdlc, self.ui.installed_dlc_container
        )
        i_widget.destroyed.connect(self.update_installed_page)
        i_widget.uninstalled.connect(self.append_available)
        self.ui.installed_dlc_container.layout().addWidget(i_widget)

    def append_available(self, rdlc: RareGame):
        self.ui.available_dlc_label.setVisible(False)
        self.ui.available_dlc_container.setVisible(True)
        i_widget: InstalledGameDlcWidget = self.get_installed(rdlc.app_name)
        if i_widget is not None:
            self.ui.available_dlc_container.layout().removeWidget(i_widget)
            i_widget.deleteLater()
        a_widget: AvailableGameDlcWidget = AvailableGameDlcWidget(
            self.rgame, rdlc, self.ui.available_dlc_container
        )
        a_widget.destroyed.connect(self.update_available_page)
        a_widget.installed.connect(self.append_installed)
        self.ui.available_dlc_container.layout().addWidget(a_widget)

    def update_dlcs(self, rgame: RareGame):
        self.rgame = rgame
        self.set_title.emit(self.rgame.app_title)

        for i_widget in self.list_installed():
            self.ui.installed_dlc_container.layout().removeWidget(i_widget)
            i_widget.deleteLater()

        for a_widget in self.list_available():
            self.ui.available_dlc_container.layout().removeWidget(a_widget)
            a_widget.deleteLater()

        for dlc in sorted(self.rgame.owned_dlcs, key=lambda x: x.app_title):
            if dlc.is_installed:
                self.append_installed(rdlc=dlc)
            else:
                self.append_available(rdlc=dlc)

        if not self.list_available():
            self.setCurrentWidget(self.ui.installed_dlc_page)
        if not self.list_installed():
            self.setCurrentWidget(self.ui.available_dlc_page)



from PyQt5.QtCore import QSettings
from PyQt5.QtWidgets import QGroupBox

from rare.shared import GlobalSignalsSingleton
from rare.ui.components.tabs.settings.widgets.rpc import Ui_RPCSettings


class RPCSettings(QGroupBox, Ui_RPCSettings):
    def __init__(self, parent):
        super(RPCSettings, self).__init__(parent=parent)
        self.setupUi(self)
        self.signals = GlobalSignalsSingleton()

        self.settings = QSettings()

        self.enable.setCurrentIndex(self.settings.value("rpc_enable", 0, int))
        self.enable.currentIndexChanged.connect(self.changed)

        self.show_game.setChecked((self.settings.value("rpc_name", True, bool)))
        self.show_game.stateChanged.connect(
            lambda: self.settings.setValue("rpc_game", self.show_game.isChecked())
        )

        self.show_os.setChecked((self.settings.value("rpc_os", True, bool)))
        self.show_os.stateChanged.connect(
            lambda: self.settings.setValue("rpc_os", self.show_os.isChecked())
        )

        self.show_time.setChecked((self.settings.value("rpc_time", True, bool)))
        self.show_time.stateChanged.connect(
            lambda: self.settings.setValue("rpc_time", self.show_time.isChecked())
        )

        try:
            import pypresence
        except ModuleNotFoundError:
            self.setDisabled(True)
            self.setToolTip(self.tr("Pypresence is not installed"))

    def changed(self, i):
        self.settings.setValue("rpc_enable", i)
        self.signals.discord_rpc.apply_settings.emit()



import re
import shutil
from logging import getLogger
from typing import Dict, Optional

from PyQt5.QtCore import pyqtSignal, QSettings, QSize, Qt, QMimeData, pyqtSlot, QCoreApplication
from PyQt5.QtGui import QDrag, QDropEvent, QDragEnterEvent, QDragMoveEvent, QFont, QMouseEvent
from PyQt5.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QInputDialog,
    QFrame,
    QMessageBox,
    QSizePolicy,
    QWidget,
    QScrollArea,
    QAction,
    QToolButton,
    QMenu,
)

from rare.shared import RareCore
from rare.ui.components.tabs.settings.widgets.wrapper import Ui_WrapperSettings
from rare.utils import config_helper
from rare.utils.misc import icon

logger = getLogger("WrapperSettings")

extra_wrapper_regex = {
    "proton": "\".*proton\" run",  # proton
    "mangohud": "mangohud"  # mangohud
}


class Wrapper:
    pass


class WrapperWidget(QFrame):
    update_wrapper = pyqtSignal(str, str)
    delete_wrapper = pyqtSignal(str)

    def __init__(self, text: str, show_text=None, parent=None):
        super(WrapperWidget, self).__init__(parent=parent)
        if not show_text:
            show_text = text.split()[0]

        self.setFrameShape(QFrame.StyledPanel)
        self.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)

        self.text = text
        self.setToolTip(text)

        unmanaged = show_text in extra_wrapper_regex.keys()

        text_lbl = QLabel(show_text, parent=self)
        text_lbl.setFont(QFont("monospace"))
        text_lbl.setDisabled(unmanaged)

        image_lbl = QLabel(parent=self)
        image_lbl.setPixmap(icon("mdi.drag-vertical").pixmap(QSize(20, 20)))

        edit_action = QAction("Edit", parent=self)
        edit_action.triggered.connect(self.__edit)
        delete_action = QAction("Delete", parent=self)
        delete_action.triggered.connect(self.__delete)

        manage_menu = QMenu(parent=self)
        manage_menu.addActions([edit_action, delete_action])

        manage_button = QToolButton(parent=self)
        manage_button.setIcon(icon("mdi.menu"))
        manage_button.setMenu(manage_menu)
        manage_button.setPopupMode(QToolButton.InstantPopup)
        manage_button.setDisabled(unmanaged)
        if unmanaged:
            manage_button.setToolTip(self.tr("Manage through settings"))
        else:
            manage_button.setToolTip(self.tr("Manage"))

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(image_lbl)
        layout.addWidget(text_lbl)
        layout.addWidget(manage_button)
        self.setLayout(layout)

        # lk: set object names for the stylesheet
        self.setObjectName(type(self).__name__)
        manage_button.setObjectName(f"{self.objectName()}Button")

    @pyqtSlot()
    def __delete(self):
        self.delete_wrapper.emit(self.text)

    def __edit(self) -> None:
        dialog = QInputDialog(self)
        dialog.setWindowTitle(f"{self.tr('Edit wrapper')} - {QCoreApplication.instance().applicationName()}")
        dialog.setLabelText(self.tr("Edit wrapper command"))
        dialog.setTextValue(self.text)
        accepted = dialog.exec()
        wrapper = dialog.textValue()
        dialog.deleteLater()
        if accepted and wrapper:
            self.update_wrapper.emit(self.text, wrapper)

    def mouseMoveEvent(self, a0: QMouseEvent) -> None:
        if a0.buttons() == Qt.LeftButton:
            a0.accept()
            drag = QDrag(self)
            mime = QMimeData()
            drag.setMimeData(mime)
            drag.exec_(Qt.MoveAction)


class WrapperSettings(QWidget):
    def __init__(self):
        super(WrapperSettings, self).__init__()
        self.ui = Ui_WrapperSettings()
        self.ui.setupUi(self)

        self.wrappers: Dict[str, WrapperWidget] = {}
        self.app_name: str = "default"

        self.wrapper_scroll = QScrollArea(self.ui.widget_stack)
        self.wrapper_scroll.setWidgetResizable(True)
        self.wrapper_scroll.setSizeAdjustPolicy(QScrollArea.AdjustToContents)
        self.wrapper_scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.wrapper_scroll.setProperty("no_kinetic_scroll", True)
        self.scroll_content = WrapperContainer(
            save_cb=self.save, parent=self.wrapper_scroll
        )
        self.wrapper_scroll.setWidget(self.scroll_content)
        self.ui.widget_stack.insertWidget(0, self.wrapper_scroll)

        self.core = RareCore.instance().core()

        self.ui.add_button.clicked.connect(self.add_button_pressed)
        self.settings = QSettings()

        self.wrapper_scroll.horizontalScrollBar().rangeChanged.connect(self.adjust_scrollarea)

        # lk: set object names for the stylesheet
        self.setObjectName(type(self).__name__)
        self.ui.no_wrapper_label.setObjectName(f"{self.objectName()}Label")
        self.wrapper_scroll.setObjectName(f"{self.objectName()}Scroll")
        self.wrapper_scroll.horizontalScrollBar().setObjectName(
            f"{self.wrapper_scroll.objectName()}Bar")
        self.wrapper_scroll.verticalScrollBar().setObjectName(
            f"{self.wrapper_scroll.objectName()}Bar")

    @pyqtSlot(int, int)
    def adjust_scrollarea(self, min: int, max: int):
        wrapper_widget = self.scroll_content.findChild(WrapperWidget)
        if not wrapper_widget:
             return
        # lk: when the scrollbar is not visible, min and max are 0
        if max > min:
            self.wrapper_scroll.setMaximumHeight(
                wrapper_widget.sizeHint().height()
                + self.wrapper_scroll.rect().height() // 2
                - self.wrapper_scroll.contentsRect().height() // 2
                + self.scroll_content.layout().spacing()
                + self.wrapper_scroll.horizontalScrollBar().sizeHint().height()
            )
        else:
            self.wrapper_scroll.setMaximumHeight(
                wrapper_widget.sizeHint().height()
                + self.wrapper_scroll.rect().height()
                - self.wrapper_scroll.contentsRect().height()
            )

    def get_wrapper_string(self):
        return " ".join(self.get_wrapper_list())

    def get_wrapper_list(self):
        wrappers = list(self.wrappers.values())
        wrappers.sort(key=lambda x: self.scroll_content.layout().indexOf(x))
        return [w.text for w in wrappers]

    def add_button_pressed(self):
        dialog = QInputDialog(self)
        dialog.setWindowTitle(f"{self.tr('Add wrapper')} - {QCoreApplication.instance().applicationName()}")
        dialog.setLabelText(self.tr("Enter wrapper command"))
        accepted = dialog.exec()
        wrapper = dialog.textValue()
        dialog.deleteLater()
        if accepted:
            self.add_wrapper(wrapper)

    def add_wrapper(self, text: str, position: int = -1, from_load: bool = False):
        if text == "mangohud" and self.wrappers.get("mangohud"):
            return
        show_text = ""
        for key, extra_wrapper in extra_wrapper_regex.items():
            if re.match(extra_wrapper, text):
                show_text = key
        if not show_text:
            show_text = text.split()[0]

        # validate
        if not text.strip():  # is empty
            return
        if not from_load:
            if self.wrappers.get(text):
                QMessageBox.warning(
                    self, self.tr("Warning"), self.tr("Wrapper <b>{0}</b> is already in the list").format(text)
                )
                return

            if show_text != "proton" and not shutil.which(text.split()[0]):
                if (
                    QMessageBox.question(
                        self,
                        self.tr("Warning"),
                        self.tr("Wrapper <b>{0}</b> is not in $PATH. Add it anyway?").format(show_text),
                        QMessageBox.Yes | QMessageBox.No,
                        QMessageBox.No,
                    )
                    == QMessageBox.No
                ):
                    return

            if text == "proton":
                QMessageBox.warning(
                    self,
                    self.tr("Warning"),
                    self.tr("Do not insert <b>proton</b> manually. Add it through Proton settings"),
                )
                return

        self.ui.widget_stack.setCurrentIndex(0)

        if widget := self.wrappers.get(show_text, None):
            widget.deleteLater()

        widget = WrapperWidget(text, show_text, self.scroll_content)
        if position < 0:
            self.scroll_content.layout().addWidget(widget)
        else:
            self.scroll_content.layout().insertWidget(position, widget)
        self.adjust_scrollarea(
            self.wrapper_scroll.horizontalScrollBar().minimum(),
            self.wrapper_scroll.horizontalScrollBar().maximum(),
        )
        widget.update_wrapper.connect(self.update_wrapper)
        widget.delete_wrapper.connect(self.delete_wrapper)

        self.wrappers[show_text] = widget

        if not from_load:
            self.save()

    @pyqtSlot(str)
    def delete_wrapper(self, text: str):
        text = text.split()[0]
        widget = self.wrappers.get(text, None)
        if widget:
            self.wrappers.pop(text)
            widget.deleteLater()

        if not self.wrappers:
            self.wrapper_scroll.setMaximumHeight(self.ui.label_page.sizeHint().height())
            self.ui.widget_stack.setCurrentIndex(1)

        self.save()

    @pyqtSlot(str, str)
    def update_wrapper(self, old: str, new: str):
        key = old.split()[0]
        idx = self.scroll_content.layout().indexOf(self.wrappers[key])
        self.delete_wrapper(key)
        self.add_wrapper(new, position=idx)

    def save(self):
        # save wrappers twice, to support wrappers with spaces
        if len(self.wrappers) == 0:
            config_helper.remove_option(self.app_name, "wrapper")
            self.settings.remove(f"{self.app_name}/wrapper")
        else:
            config_helper.add_option(self.app_name, "wrapper", self.get_wrapper_string())
            self.settings.setValue(f"{self.app_name}/wrapper", self.get_wrapper_list())

    def load_settings(self, app_name: str):
        self.app_name = app_name
        for i in self.wrappers.values():
            i.deleteLater()
        self.wrappers.clear()

        wrappers = self.settings.value(f"{self.app_name}/wrapper", [], str)

        if not wrappers and (cfg := self.core.lgd.config.get(self.app_name, "wrapper", fallback="")):
            logger.info("Loading wrappers from legendary config")
            # no qt wrapper, but legendary wrapper, to have backward compatibility
            pattern = re.compile(r'''((?:[^ "']|"[^"]*"|'[^']*')+)''')
            wrappers = pattern.split(cfg)[1::2]

        for wrapper in wrappers:
            self.add_wrapper(wrapper, from_load=True)

        if not self.wrappers:
            self.wrapper_scroll.setMaximumHeight(self.ui.label_page.sizeHint().height())
            self.ui.widget_stack.setCurrentIndex(1)
        else:
            self.ui.widget_stack.setCurrentIndex(0)

        self.save()


class WrapperContainer(QWidget):

    def __init__(self, save_cb, parent=None):
        super(WrapperContainer, self).__init__(parent=parent)
        self.setAcceptDrops(True)
        self.save = save_cb
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.setLayout(layout)

        self.drag_widget: Optional[QWidget] = None

        # lk: set object names for the stylesheet
        self.setObjectName(type(self).__name__)

    def dragEnterEvent(self, e: QDragEnterEvent):
        widget = e.source()
        self.drag_widget = widget
        e.accept()

    def _get_drop_index(self, x):
        drag_idx = self.layout().indexOf(self.drag_widget)

        if drag_idx > 0:
            prev_widget = self.layout().itemAt(drag_idx - 1).widget()
            if x < self.drag_widget.x() - prev_widget.width() // 2:
                return drag_idx - 1
        if drag_idx < self.layout().count() - 1:
            next_widget = self.layout().itemAt(drag_idx + 1).widget()
            if x > self.drag_widget.x() + self.drag_widget.width() + next_widget.width() // 2:
                return drag_idx + 1

        return drag_idx

    def dragMoveEvent(self, e: QDragMoveEvent) -> None:
        i = self._get_drop_index(e.pos().x())
        self.layout().insertWidget(i, self.drag_widget)

    def dropEvent(self, e: QDropEvent):
        pos = e.pos()
        widget = e.source()
        index = self._get_drop_index(pos.x())
        self.layout().insertWidget(index, widget)
        self.drag_widget = None
        e.accept()
        self.save()



from PyQt5.QtCore import QCoreApplication

from .overlay_settings import OverlaySettings, CustomOption


class DxvkSettings(OverlaySettings):
    def __init__(self):
        super(DxvkSettings, self).__init__(
            [
                ("fps", QCoreApplication.translate("DxvkSettings", "FPS")),
                ("frametime", QCoreApplication.translate("DxvkSettings", "Frametime")),
                ("memory", QCoreApplication.translate("DxvkSettings", "Memory usage")),
                ("gpuload", QCoreApplication.translate("DxvkSettings", "GPU usage")),
                ("devinfo", QCoreApplication.translate("DxvkSettings", "Show Device info")),
                ("version", QCoreApplication.translate("DxvkSettings", "DXVK Version")),
                ("api", QCoreApplication.translate("DxvkSettings", "D3D feature level")),
                ("compiler", QCoreApplication.translate("DxvkSettings", "Compiler activity")),
            ],
            [
                (CustomOption.number_input("scale", 1, True), QCoreApplication.translate("DxvkSettings", "Scale"))
            ],
            "DXVK_HUD", "0"
        )

        self.setTitle(self.tr("DXVK Settings"))
        self.gb_options.setTitle(self.tr("Custom options"))



import sys
from abc import abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from typing import Optional

from PyQt5.QtCore import QRunnable, QObject, pyqtSlot, pyqtSignal


class Worker(QRunnable):
    """
    Base QRunnable class.

    This class provides a base for QRunnables with signals that are automatically deleted.

    To use this class you have to assign the signals object of your concrete implementation
    to the `Worker.signals` attribute and implement `Worker.run_real()`
    """

    def __init__(self):
        super(Worker, self).__init__()
        self.setAutoDelete(True)
        self.__signals: Optional[QObject] = None

    @property
    def signals(self) -> QObject:
        if self.__signals is None:
            raise RuntimeError(f"Subclasses must assign '{type(self).__name__}.signals' QObject attribute")
        return self.__signals

    @signals.setter
    def signals(self, obj: QObject):
        self.__signals = obj

    @abstractmethod
    def run_real(self):
        pass

    @pyqtSlot()
    def run(self):
        self.run_real()
        self.signals.deleteLater()


class QueueWorkerState(IntEnum):
    UNDEFINED = 0
    QUEUED = 1
    ACTIVE = 2


@dataclass
class QueueWorkerInfo:
    app_name: str
    app_title: str
    worker_type: str
    state: QueueWorkerState
    progress: int = 0


class QueueWorker(Worker):
    """
    Base queueable worker class

    This class is a specialization of the `Worker` class. It provides feedback signals to know
    if a worker has started or finished.

    To use this class you have to assign the signals object of your concrete implementation
    to the `QueueWorker.signals` attribute, implement `QueueWorker.run_real()` and `QueueWorker.worker_info()`
    """

    class Signals(QObject):
        started = pyqtSignal()
        finished = pyqtSignal()

    def __init__(self):
        super(QueueWorker, self).__init__()
        self.feedback = QueueWorker.Signals()
        self.state = QueueWorkerState.QUEUED
        self._kill = False

    @pyqtSlot()
    def run(self):
        self.state = QueueWorkerState.ACTIVE
        self.feedback.started.emit()
        super(QueueWorker, self).run()
        self.feedback.finished.emit()
        self.feedback.deleteLater()

    @abstractmethod
    def worker_info(self) -> QueueWorkerInfo:
        pass

    def kill(self):
        raise NotImplementedError
        self._kill = True




import os
import platform
import time
from configparser import ConfigParser
from logging import getLogger
from typing import Union, Iterable

from PyQt5.QtCore import pyqtSignal, QObject, QRunnable

import rare.utils.wine as wine
from rare.lgndr.core import LegendaryCore
from rare.models.game import RareGame
from rare.models.pathspec import PathSpec
from rare.utils.misc import path_size, format_size
from .worker import Worker

if platform.system() == "Windows":
    # noinspection PyUnresolvedReferences
    import winreg # pylint: disable=E0401
    from legendary.lfs import windows_helpers

logger = getLogger("WineResolver")


class WineResolver(Worker):
    class Signals(QObject):
        result_ready = pyqtSignal(str)

    def __init__(self, core: LegendaryCore, path: str, app_name: str):
        super(WineResolver, self).__init__()
        self.signals = WineResolver.Signals()
        self.wine_env = wine.environ(core, app_name)
        self.wine_exec = wine.wine(core, app_name)
        self.path = PathSpec(core, app_name).cook(path)

    def run_real(self):
        if "WINEPREFIX" not in self.wine_env or not os.path.exists(self.wine_env["WINEPREFIX"]):
            # pylint: disable=E1136
            self.signals.result_ready[str].emit("")
            return
        if not os.path.exists(self.wine_exec):
            # pylint: disable=E1136
            self.signals.result_ready[str].emit("")
            return
        path = wine.resolve_path(self.wine_exec, self.wine_env, self.path)
        # Clean wine output
        real_path = wine.convert_to_unix_path(self.wine_exec, self.wine_env, path)
        # pylint: disable=E1136
        self.signals.result_ready[str].emit(real_path)
        return


class OriginWineWorker(QRunnable):
    def __init__(self, core: LegendaryCore, games: Union[Iterable[RareGame], RareGame]):
        super(OriginWineWorker, self).__init__()
        self.__cache: dict[str, ConfigParser] = {}
        self.core = core
        if isinstance(games, RareGame):
            games = [games]
        self.games = games

    def run(self) -> None:
        t = time.time()

        for rgame in self.games:

            reg_path: str = rgame.game.metadata \
                .get("customAttributes", {}) \
                .get("RegistryPath", {}).get("value", None)
            if not reg_path:
                continue

            reg_key: str = rgame.game.metadata \
                .get("customAttributes", {}) \
                .get("RegistryKey", {}).get("value", None)
            if not reg_key:
                continue

            if platform.system() == "Windows":
                install_dir = windows_helpers.query_registry_value(winreg.HKEY_LOCAL_MACHINE, reg_path, reg_key)
            else:
                wine_env = wine.environ(self.core, rgame.app_name)
                wine_exec = wine.wine(self.core, rgame.app_name)

                use_wine = False
                if not use_wine:
                    # lk: this is the original way of getting the path by parsing "system.reg"
                    wine_prefix = wine.prefix(self.core, rgame.app_name)
                    reg = self.__cache.get(wine_prefix, None) or wine.read_registry("system.reg", wine_prefix)
                    self.__cache[wine_prefix] = reg

                    reg_path = reg_path.replace("SOFTWARE", "Software").replace("WOW6432Node", "Wow6432Node")
                    # lk: split and rejoin the registry path to avoid slash expansion
                    reg_path = "\\\\".join([x for x in reg_path.split("\\") if bool(x)])

                    install_dir = reg.get(reg_path, f'"{reg_key}"', fallback=None)
                else:
                    # lk: this is the alternative way of getting the path by using wine itself
                    install_dir = wine.query_reg_key(wine_exec, wine_env, f"HKLM\\{reg_path}", reg_key)

                if install_dir:
                    logger.debug("Found Wine install directory %s", install_dir)
                    install_dir = wine.convert_to_unix_path(wine_exec, wine_env, install_dir)
                    if install_dir:
                        logger.debug("Found Unix install directory %s", install_dir)
                    else:
                        logger.info("Could not find Unix install directory for %s", rgame.app_title)
                else:
                    logger.info("Could not find Wine install directory for %s", rgame.app_title)

            if install_dir:
                if os.path.isdir(install_dir):
                    install_size = path_size(install_dir)
                    rgame.set_origin_attributes(install_dir, install_size)
                    logger.info("Origin game %s (%s, %s)", rgame.app_title, install_dir, format_size(install_size))
                else:
                    logger.warning("Origin game %s (%s does not exist)", rgame.app_title, install_dir)
            else:
                logger.info("Origin game %s is not installed", rgame.app_title)
        logger.info("Origin worker finished in %ss", time.time() - t)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/games/game_info/game_dlc.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_GameDlc(object):
    def setupUi(self, GameDlc):
        GameDlc.setObjectName("GameDlc")
        GameDlc.resize(271, 139)
        GameDlc.setWindowTitle("GameDlc")
        GameDlc.setFrameShape(QtWidgets.QFrame.StyledPanel)
        GameDlc.setFrameShadow(QtWidgets.QFrame.Sunken)
        GameDlc.setLineWidth(0)
        self.installed_dlc_page = QtWidgets.QWidget()
        self.installed_dlc_page.setGeometry(QtCore.QRect(0, 0, 267, 79))
        self.installed_dlc_page.setObjectName("installed_dlc_page")
        self.installed_dlc_page_layout = QtWidgets.QVBoxLayout(self.installed_dlc_page)
        self.installed_dlc_page_layout.setContentsMargins(0, 0, 0, 0)
        self.installed_dlc_page_layout.setObjectName("installed_dlc_page_layout")
        self.installed_dlc_label = QtWidgets.QLabel(self.installed_dlc_page)
        self.installed_dlc_label.setObjectName("installed_dlc_label")
        self.installed_dlc_page_layout.addWidget(self.installed_dlc_label, 0, QtCore.Qt.AlignTop)
        self.installed_dlc_container = QtWidgets.QWidget(self.installed_dlc_page)
        self.installed_dlc_container.setObjectName("installed_dlc_container")
        self.installed_dlc_container_layout = QtWidgets.QVBoxLayout(self.installed_dlc_container)
        self.installed_dlc_container_layout.setContentsMargins(0, 0, 3, 0)
        self.installed_dlc_container_layout.setObjectName("installed_dlc_container_layout")
        self.installed_dlc_page_layout.addWidget(self.installed_dlc_container, 0, QtCore.Qt.AlignTop)
        self.installed_dlc_page_layout.setStretch(1, 1)
        GameDlc.addItem(self.installed_dlc_page, "")
        self.available_dlc_page = QtWidgets.QWidget()
        self.available_dlc_page.setGeometry(QtCore.QRect(0, 0, 267, 79))
        self.available_dlc_page.setObjectName("available_dlc_page")
        self.available_dlc_page_layou = QtWidgets.QVBoxLayout(self.available_dlc_page)
        self.available_dlc_page_layou.setContentsMargins(0, 0, 0, 0)
        self.available_dlc_page_layou.setObjectName("available_dlc_page_layou")
        self.available_dlc_label = QtWidgets.QLabel(self.available_dlc_page)
        self.available_dlc_label.setObjectName("available_dlc_label")
        self.available_dlc_page_layou.addWidget(self.available_dlc_label, 0, QtCore.Qt.AlignTop)
        self.available_dlc_container = QtWidgets.QWidget(self.available_dlc_page)
        self.available_dlc_container.setObjectName("available_dlc_container")
        self.available_dlc_container_layout = QtWidgets.QVBoxLayout(self.available_dlc_container)
        self.available_dlc_container_layout.setContentsMargins(0, 0, 3, 0)
        self.available_dlc_container_layout.setObjectName("available_dlc_container_layout")
        self.available_dlc_page_layou.addWidget(self.available_dlc_container, 0, QtCore.Qt.AlignTop)
        self.available_dlc_page_layou.setStretch(1, 1)
        GameDlc.addItem(self.available_dlc_page, "")

        self.retranslateUi(GameDlc)

    def retranslateUi(self, GameDlc):
        _translate = QtCore.QCoreApplication.translate
        self.installed_dlc_label.setText(_translate("GameDlc", "No Downloadable Content has been installed."))
        GameDlc.setItemText(GameDlc.indexOf(self.installed_dlc_page), _translate("GameDlc", "Installed DLCs"))
        self.available_dlc_label.setText(_translate("GameDlc", "No Downloadable Content is available"))
        GameDlc.setItemText(GameDlc.indexOf(self.available_dlc_page), _translate("GameDlc", "Available DLCs"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    GameDlc = QtWidgets.QToolBox()
    ui = Ui_GameDlc()
    ui.setupUi(GameDlc)
    GameDlc.show()
    sys.exit(app.exec_())
